from twisted.python.constants import Names, NamedConstant
from twisted.trial.unittest import TestCase

from vumi.transports.parlayx.xmlutil import (
    Namespace, QualifiedName, ElementMaker, LocalNamespace as L,
    split_qualified, gettext, gettextall, tostring, elemfind, elemfindall,
    element_to_dict)


class NamespaceTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.xmlutil.Namespace`.
    """
    def test_str(self):
        """
        ``str(Namespace)`` produces the Namespace URI.
        """
        uri = 'http://example.com'
        self.assertEqual(uri, str(Namespace(uri)))

    def test_repr(self):
        """
        ``repr(Namespace)`` produces self-explanatory human-readable output.
        """
        self.assertEqual(
            '<Namespace uri=None prefix=None>',
            repr(Namespace(None)))
        self.assertEqual(
            "<Namespace uri='http://example.com' prefix=None>",
            repr(Namespace('http://example.com')))
        self.assertEqual(
            "<Namespace uri='http://example.com' prefix='ex'>",
            repr(Namespace('http://example.com', 'ex')))

    def test_equality(self):
        """
        Two `Namespace` instances created with the same values compare equal to
        one another.
        """
        self.assertEqual(
            Namespace('http://example.com'),
            Namespace('http://example.com'))
        self.assertEqual(
            Namespace('http://example.com', 'ex'),
            Namespace('http://example.com', 'ex'))
        self.assertNotEqual(
            Namespace('http://example.com'),
            Namespace('http://example.com', 'ex'))
        self.assertNotEqual(
            Namespace('http://example.com/'),
            Namespace('http://example.com'))

    def test_qualified_name(self):
        """
        `Namespace.__getattr__` produces qualified `QualifiedName` instances if
        `Namespace.__uri` is not `None`.
        """
        uri = 'http://example.com'
        ns = Namespace(uri)
        self.assertEqual(
            QualifiedName(uri, 'foo'),
            ns.foo)

    def test_local_name(self):
        """
        `Namespace.__getattr__` produces local `QualifiedName` instances if
        `Namespace.__uri` is `None`.
        """
        ns = Namespace(None)
        self.assertEqual(
            QualifiedName('foo'),
            ns.foo)


class QualifiedNameTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.xmlutil.QualifiedName`.
    """
    def test_repr(self):
        """
        ``repr(QualifiedName)`` produces self-explanatory human-readable
        output.
        """
        self.assertEqual(
            "<QualifiedName xmlns=None local='tag'>",
            repr(QualifiedName('tag')))
        self.assertEqual(
            "<QualifiedName xmlns='http://example.com' local='tag'>",
            repr(QualifiedName('http://example.com', 'tag')))

    def test_equality(self):
        """
        Two `QualifiedName` instances created with the same values compare
        equal to one another.
        """
        self.assertEqual(
            QualifiedName('tag'),
            QualifiedName('tag'))
        self.assertEqual(
            QualifiedName('http://example.com', 'tag'),
            QualifiedName('http://example.com', 'tag'))
        # Parameters are internally converted to Clark notation anyway.
        self.assertEqual(
            QualifiedName('http://example.com', 'tag'),
            QualifiedName('{http://example.com}tag'))
        self.assertNotEqual(
            QualifiedName('tag'),
            QualifiedName('http://example.com', 'tag'))
        self.assertNotEqual(
            QualifiedName('http://example.com/', 'tag'),
            QualifiedName('http://example.com', 'tag'))

    def test_element(self):
        """
        `QualifiedName` instances are callable and produce ElementTree
        elements.
        """
        qname = QualifiedName('tag')
        self.assertEqual(
            '<tag />',
            tostring(qname()))
        self.assertEqual(
            '<tag>hello</tag>',
            tostring(qname(u'hello')))
        self.assertEqual(
            '<tag key="value">hello</tag>',
            tostring(qname('hello', key='value')))
        self.assertEqual(
            '<tag key="value">hello</tag>',
            tostring(qname('hello', dict(key='value'))))


class ElementMakerTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.xmlutil.ElementMaker`.
    """
    def setUp(self):
        # ElementTree has a global namespace prefix map. We need to patch it
        # out here to make the tests independent of each other.
        import xml.etree.ElementTree
        self.patch(xml.etree.ElementTree, '_namespace_map', {})

    def test_unknown_child_type(self):
        """
        `ElementMaker` instances raise `TypeError` when called with children of
        unmapped types.
        """
        E = ElementMaker()
        exc = self.assertRaises(TypeError, E, 'tag', None)
        self.assertEqual('Unknown child type: None', str(exc))

    def test_simple(self):
        """
        Calling `ElementMaker` instances produces ElementTree elements,
        children and attributes can be provided too.
        """
        E = ElementMaker()
        self.assertEqual(
            '<tag />',
            tostring(E('tag')))
        self.assertEqual(
            '<tag>hello</tag>',
            tostring(E('tag', 'hello')))
        self.assertEqual(
            '<tag key="value">hello</tag>',
            tostring(E('tag', 'hello', key='value')))
        self.assertEqual(
            '<tag key="value">hello</tag>',
            tostring(E('tag', 'hello', dict(key='value'))))

    def test_callable(self):
        """
        Providing a callable child will result in that child being called, with
        no arguments, to produce the actual child value.
        """
        E = ElementMaker()
        self.assertEqual(
            '<tag><child /></tag>',
            tostring(E('tag', L.child)))
        self.assertEqual(
            '<tag>hello</tag>',
            tostring(E('tag', lambda: 'hello')))

    def test_list(self):
        """
        Providing a list child will result in all the elements of the list
        added individually.
        """
        E = ElementMaker()
        self.assertEqual(
            '<tag><child1 /><child2 /></tag>',
            tostring(E('tag', [L.child1, L.child2])))
        self.assertEqual(
            '<tag>text1text2</tag>',
            tostring(E('tag', ['text1', 'text2'])))

    def test_nested(self):
        """
        Children can themselves be ElementTree elements, resulting in nested
        elements.
        """
        E = ElementMaker()
        self.assertEqual(
            '<tag><child /></tag>',
            tostring(E('tag', E('child'))))
        self.assertEqual(
            '<tag><child>hello</child></tag>',
            tostring(E('tag', E('child', 'hello'))))
        self.assertEqual(
            '<tag><child key="value">hello</child></tag>',
            tostring(E('tag', E('child', 'hello', key='value'))))
        self.assertEqual(
            '<tag><child key="value">hello</child></tag>',
            tostring(E('tag', E('child', 'hello', dict(key='value')))))

    def test_namespaced(self):
        """
        Tags that are `QualifiedName` instances or use Clark notation produce
        namespaced XML elements.
        """
        E = ElementMaker()
        self.assertEqual(
            '<ns0:tag xmlns:ns0="http://example.com" />',
            tostring(E('{http://example.com}tag')))
        self.assertEqual(
            '<ns0:tag xmlns:ns0="http://example.com" />',
            tostring(QualifiedName('http://example.com', 'tag')()))
        ns = Namespace('http://example.com', 'ex')
        self.assertEqual(
            '<ex:tag xmlns:ex="http://example.com" />',
            tostring(ns.tag()))

    def test_namespaced_attributes(self):
        """
        XML attributes that are `QualifiedName` instances or use Clark notation
        produce namespaced XML element attributes.
        """
        ns = Namespace('http://example.com', 'ex')
        attrib = {ns.key: 'value'}
        self.assertEqual(
            {'{http://example.com}tag': {
                '@{http://example.com}key': 'value'}},
            element_to_dict(ns.tag(attrib)))
        attrib = {'{http://example.com}key': 'value'}
        self.assertEqual(
            {'{http://example.com}tag': {
                '@{http://example.com}key': 'value'}},
            element_to_dict(ns.tag(attrib)))

    def test_typemap(self):
        """
        Providing a type map to `ElementMaker` allows the caller to specify how
        to serialize types other than strings and dictionaries.
        """
        E = ElementMaker(typemap={
            float: lambda e, v: '%0.2f' % (v,),
            int: lambda e, v: L.int(str(v))})
        self.assertEqual(
            '<tag>2.50</tag>',
            tostring(E('tag', 2.5)))
        self.assertEqual(
            '<tag><int>42</int></tag>',
            tostring(E('tag', 42)))


class MetasyntacticVariables(Names):
    """
    Metasyntactic variable names.
    """
    Foo = NamedConstant()


class GetTextTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.xmlutil.gettext`.
    """
    def setUp(self):
        self.root = L.top(
            L.a('hello'),
            L.b('42'),
            L.b('24'),
            L.c('Foo'),
            L.d,
            L.sub(
                L.e('world'),
                L.e('all'),
                L.f))

    def test_simple(self):
        """
        Getting a sub-element with a `text` attribute returns the text as a
        `unicode` object.
        """
        res = gettext(self.root, u'a')
        self.assertIdentical(unicode, type(res))
        self.assertEqual(res, u'hello')

        res = gettext(self.root, u'sub/e')
        self.assertIdentical(unicode, type(res))
        self.assertEqual(res, u'world')

    def test_default(self):
        """
        Getting a sub-element without a `text` attribute, or attempting to get
        a sub-element that does not exist, results in the `default` parameter
        to `gettext` being used, defaulting to `None`.
        """
        self.assertIdentical(gettext(self.root, u'd'), None)
        self.assertEqual(gettext(self.root, u'd', default=42), 42)

        self.assertIdentical(gettext(self.root, u'sub/f'), None)
        res = gettext(self.root, u'sub/f', default='a')
        self.assertIdentical(str, type(res))
        self.assertEqual(res, 'a')

        self.assertIdentical(gettext(self.root, u'haha_what'), None)
        self.assertEqual(gettext(self.root, u'haha_what', default=42), 42)

    def test_parse(self):
        """
        Specifying a `parse` callable results in that being called to transform
        the element text.
        """
        self.assertEqual(
            42,
            gettext(self.root, u'b', parse=int))
        self.assertEqual(
            MetasyntacticVariables.Foo,
            gettext(self.root, u'c',
                    parse=MetasyntacticVariables.lookupByName))
        self.assertRaises(ValueError,
            gettext, self.root, u'c', parse=int)

    def test_parseWithDefault(self):
        """
        In the event that a default value is specified and a `parse` callable
        given, and the default value is used, the default value will be passed
        to the callable.
        """
        self.assertEqual(
            42,
            gettext(self.root, u'b', default=3.1415, parse=int))
        self.assertEqual(
            21,
            gettext(self.root, u'd', default=21, parse=int))
        self.assertRaises(ValueError,
            gettext, self.root, u'd', default='foo', parse=int)

    def test_gettextall(self):
        """
        `gettextall` is like `gettext` except it uses `elemfindall` instead of
        `elemfind`, returning a ``list`` of results.
        """
        self.assertEqual(
            [42, 24],
            list(gettextall(self.root, u'b', parse=int)))
        self.assertEqual(
            ['world', 'all'],
            list(gettextall(self.root, u'sub/e')))
        self.assertEqual(
            [],
            list(gettextall(self.root, u'what')))


class SplitQualifiedTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.xmlutil.split_qualified`.
    """
    def test_local(self):
        """
        `split_qualified` splits a local XML name into `None` and the tag name.
        """
        self.assertEqual((None, 'tag'), split_qualified('tag'))

    def test_qualified(self):
        """
        `split_qualified` splits a qualified XML name into a URI and the tag
        name.
        """
        self.assertEqual(
            ('http://example.com', 'tag'),
            split_qualified('{http://example.com}tag'))


class FindTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.xmlutil.elemfind`.
    """
    def setUp(self):
        self.root = L.parent(
            L.child1, L.child2, L.child2, L.child3)

    def test_elemfind(self):
        """
        `elemfind` finds the first `QualifiedName` or path specified.
        """
        self.assertEqual(
            '<child1 />',
            tostring(elemfind(self.root, 'child1')))
        self.assertEqual(
            '<child2 />',
            tostring(elemfind(self.root, L.child2)))

    def test_elemfind_none(self):
        """
        `elemfind` returns ``None`` if the `QualifiedName` or path specified
        cannot be found.
        """
        self.assertIdentical(None, elemfind(self.root, L.what))

    def test_elemfindall(self):
        """
        `elemfind` finds all sub-elements with the `QualifiedName` or path
        specified.
        """
        self.assertEqual(
            ['<child1 />'],
            map(tostring, elemfindall(self.root, L.child1)))
        self.assertEqual(
            ['<child2 />', '<child2 />'],
            map(tostring, elemfindall(self.root, 'child2')))

    def test_elemfindall_none(self):
        """
        `elemfind` returns an empty list if the `QualifiedName` or path
        specified cannot be found.
        """
        self.assertEqual([], elemfindall(self.root, L.what))


class ElementToDictTests(TestCase):
    """
    Tests for `vumi.transports.parlayx.xmlutil.element_to_dict`.
    """
    def test_empty(self):
        """
        An empty element produces a ``None`` value keyed against its tag name.
        """
        self.assertEqual(
            {'root': None},
            element_to_dict(L.root()))

    def test_empty_attributes(self):
        """
        An element containing only attributes, and no content, has its
        attributes, prefixed with an ``@`` keyed against its tag name.
        """
        self.assertEqual(
            {'root': {'@attr': 'value'}},
            element_to_dict(L.root(attr='value')))

    def test_text(self):
        """
        An element containing only text content, has its text keyed against its
        tag name.
        """
        self.assertEqual(
            {'root': 'hello'},
            element_to_dict(L.root('hello')))

    def test_text_attributes(self):
        """
        An element containing attributes and text content, has its
        attributes, prefixed with an ``@`` keyed against its tag name and its
        text keyed against ``#text``.
        """
        self.assertEqual(
            {'root': {'#text': 'hello', '@attr': 'value'}},
            element_to_dict(L.root('hello', attr='value')))

    def test_children_text(self):
        """
        Child elements are recursively nested.

        An element containing only text content, has its text keyed against its
        tag name.
        """
        self.assertEqual(
            {'root': {'child': 'hello'}},
            element_to_dict(
                L.root(L.child('hello'))))

    def test_children_attributes(self):
        """
        Child elements are recursively nested.

        An element containing only attributes, and no content, has its
        attributes, prefixed with an ``@`` keyed against its tag name.
        """
        self.assertEqual(
            {'root': {'child': {'@attr': 'value'}}},
            element_to_dict(
                L.root(L.child(attr='value'))))

    def test_children_text_attributes(self):
        """
        Child elements are recursively nested.

        An element containing attributes and text content, has its
        attributes, prefixed with an ``@`` keyed against its tag name and its
        text keyed against ``#text``.
        """
        self.assertEqual(
            {'root': {'child': {'#text': 'hello', '@attr': 'value'}}},
            element_to_dict(L.root(L.child('hello', attr='value'))))

    def test_children_multiple(self):
        """
        Multiple child elements with the same tag name are coalesced into
        a ``list``.
        """
        self.assertEqual(
            {'root': {'child': [{'@attr': 'value'}, 'hello']}},
            element_to_dict(
                L.root(L.child(attr='value'), L.child('hello'))))

    def test_namespaced(self):
        """
        `element_to_dict` supports namespaced element names and namespaced
        attributes.
        """
        ns = Namespace('http://example.com', 'ex')
        self.assertEqual(
            {str(ns.root): {
                'child': [
                    {'@' + str(ns.attr): 'value'},
                    {'@attr2': 'value2',
                     '#text': 'hello'},
                    'world']}},
            element_to_dict(
                ns.root(
                    L.child({ns.attr: 'value'}),
                    L.child('hello', attr2='value2'),
                    L.child('world'))))



# -*- test-case-name: vumi.transports.parlayx.tests.test_server -*-
import iso8601
from collections import namedtuple

from twisted.internet.defer import maybeDeferred, fail
from twisted.python import log
from twisted.python.constants import Values, ValueConstant
from twisted.web import http
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET

from vumi.transports.parlayx.client import PARLAYX_COMMON_NS
from vumi.transports.parlayx.soaputil import (
    soap_envelope, unwrap_soap_envelope, soap_fault, SoapFault)
from vumi.transports.parlayx.xmlutil import (
    Namespace, elemfind, gettext, split_qualified, parse_document, tostring)
from vumi.utils import normalize_msisdn


NOTIFICATION_NS = Namespace(
    'http://www.csapi.org/schema/parlayx/sms/notification/v2_2/local', 'loc')


def normalize_address(address):
    """
    Normalize a ParlayX address.
    """
    if address.startswith('tel:'):
        address = address[4:]
    return normalize_msisdn(address)


class DeliveryStatus(Values):
    """
    ParlayX `DeliveryStatus` enumeration type.
    """
    DeliveredToNetwork = ValueConstant('delivered')
    DeliveryUncertain = ValueConstant('pending')
    DeliveryImpossible = ValueConstant('failed')
    MessageWaiting = ValueConstant('pending')
    DeliveredToTerminal = ValueConstant('delivered')
    DeliveryNotificationNotSupported = ValueConstant('failed')


class SmsMessage(namedtuple('SmsMessage',
                            ['message', 'sender_address',
                             'service_activation_number', 'timestamp'])):
    """
    ParlayX `SmsMessage` complex type.
    """
    @classmethod
    def from_element(cls, root):
        """
        Create an `SmsMessage` instance from an ElementTree element.
        """
        return cls(
            message=gettext(root, 'message'),
            sender_address=gettext(
                root, 'senderAddress', parse=normalize_address),
            service_activation_number=gettext(
                root, 'smsServiceActivationNumber', parse=normalize_address),
            timestamp=gettext(root, 'dateTime', parse=iso8601.parse_date))


class DeliveryInformation(namedtuple('DeliveryInformation',
                                     ['address', 'delivery_status'])):
    """
    ParlayX `DeliveryInformation` complex type.
    """
    @classmethod
    def from_element(cls, root):
        """
        Create a `DeliveryInformation` instance from an ElementTree element.
        """
        try:
            delivery_status = gettext(
                root, 'deliveryStatus', parse=DeliveryStatus.lookupByName)
        except ValueError, e:
            raise ValueError(
                'No such delivery status enumeration value: %r' % (str(e),))
        else:
            return cls(
                address=gettext(root, 'address', parse=normalize_address),
                delivery_status=delivery_status)


class SmsNotificationService(Resource):
    """
    Web resource to handle SOAP requests for ParlayX SMS deliveries and
    delivery receipts.
    """
    isLeaf = True

    def __init__(self, callback_message_received, callback_message_delivered):
        self.callback_message_received = callback_message_received
        self.callback_message_delivered = callback_message_delivered
        Resource.__init__(self)

    def render_POST(self, request):
        """
        Process a SOAP request and convert any exceptions into SOAP faults.
        """
        def _writeResponse(response):
            request.setHeader('Content-Type', 'text/xml; charset="utf-8"')
            request.write(tostring(soap_envelope(response)))
            request.finish()

        def _handleSuccess(result):
            request.setResponseCode(http.OK)
            return result

        def _handleError(f):
            # XXX: Perhaps report this back to the transport somehow???
            log.err(f, 'Failure processing SOAP request')
            request.setResponseCode(http.INTERNAL_SERVER_ERROR)
            faultcode = u'soapenv:Server'
            if f.check(SoapFault):
                return f.value.to_element()
            return soap_fault(faultcode, f.getErrorMessage())

        try:
            tree = parse_document(request.content)
            body, header = unwrap_soap_envelope(tree)
        except:
            d = fail()
        else:
            d = maybeDeferred(self.process, request, body, header)
            d.addCallback(_handleSuccess)

        d.addErrback(_handleError)
        d.addCallback(_writeResponse)
        return NOT_DONE_YET

    def process(self, request, body, header=None):
        """
        Process a SOAP request.
        """
        for child in body.getchildren():
            # Since there is no SOAPAction header, and these requests are not
            # made to different endpoints, the only way to handle these is to
            # switch on the root element's name. Yuck.
            localname = split_qualified(child.tag)[1]
            meth = getattr(self, 'process_' + localname, self.process_unknown)
            return meth(child, header, localname)

        raise SoapFault(u'soapenv:Client', u'No actionable items')

    def process_unknown(self, root, header, name):
        """
        Process unknown notification deliverables.
        """
        raise SoapFault(u'soapenv:Server', u'No handler for %s' % (name,))

    def process_notifySmsReception(self, root, header, name):
        """
        Process a received text message.
        """
        linkid = None
        if header is not None:
            linkid = gettext(header, './/' + str(PARLAYX_COMMON_NS.linkid))

        correlator = gettext(root, NOTIFICATION_NS.correlator)
        message = SmsMessage.from_element(
            elemfind(root, NOTIFICATION_NS.message))
        d = maybeDeferred(
            self.callback_message_received, correlator, linkid, message)
        d.addCallback(
            lambda ignored: NOTIFICATION_NS.notifySmsReceptionResponse())
        return d

    def process_notifySmsDeliveryReceipt(self, root, header, name):
        """
        Process a text message delivery receipt.
        """
        correlator = gettext(root, NOTIFICATION_NS.correlator)
        delivery_info = DeliveryInformation.from_element(
            elemfind(root, NOTIFICATION_NS.deliveryStatus))
        d = maybeDeferred(self.callback_message_delivered,
            correlator, delivery_info.delivery_status.value)
        d.addCallback(
            lambda ignored: NOTIFICATION_NS.notifySmsDeliveryReceiptResponse())
        return d


# XXX: Only used for debugging with SoapUI:
# twistd web --class=vumi.transports.parlayx.server.Root --port=9080
class Root(Resource):
    def getChild(self, path, request):
        from twisted.internet.defer import succeed
        noop = lambda *a, **kw: succeed(None)
        if request.postpath == ['services', 'SmsNotification']:
            return SmsNotificationService(noop, noop)
        return None


__all__ = [
    'normalize_address', 'DeliveryStatus', 'SmsMessage', 'DeliveryInformation',
    'SmsNotificationService']



# -*- coding: utf-8 -*-

"""Tests for vumi.transports.dmark.dmark_ussd."""

import json

import urllib

from twisted.internet.defer import inlineCallbacks
from twisted.internet.task import Clock

from vumi.message import TransportUserMessage
from vumi.tests.helpers import VumiTestCase
from vumi.transports.dmark import DmarkUssdTransport
from vumi.transports.httprpc.tests.helpers import HttpRpcTransportHelper


class TestDmarkUssdTransport(VumiTestCase):

    _transaction_id = u'transaction-123'
    _to_addr = '*121#'
    _from_addr = '+00775551122'
    _request_defaults = {
        'transactionId': _transaction_id,
        'msisdn': _from_addr,
        'ussdServiceCode': _to_addr,
        'transactionTime': '1389971940',
        'ussdRequestString': _to_addr,
        'creationTime': '1389971950',
        'response': 'false',
    }

    @inlineCallbacks
    def setUp(self):
        self.clock = Clock()
        self.patch(DmarkUssdTransport, 'get_clock', lambda _: self.clock)

        self.config = {
            'web_port': 0,
            'web_path': '/api/v1/dmark/ussd/',
            'publish_status': True,
        }
        self.tx_helper = self.add_helper(
            HttpRpcTransportHelper(DmarkUssdTransport,
                                   request_defaults=self._request_defaults))
        self.transport = yield self.tx_helper.get_transport(self.config)
        self.session_manager = self.transport.session_manager
        self.transport_url = self.transport.get_transport_url(
            self.config['web_path'])
        yield self.session_manager.redis._purge_all()  # just in case
        self.session_timestamps = {}

    @inlineCallbacks
    def mk_session(self, transaction_id=_transaction_id):
        yield self.session_manager.create_session(
            transaction_id, transaction_id=transaction_id)

    def assert_inbound_message(self, msg, **field_values):
        expected_field_values = {
            'content': self._request_defaults['ussdRequestString'],
            'to_addr': self._to_addr,
            'from_addr': self._from_addr,
            'session_event': TransportUserMessage.SESSION_NEW,
            'transport_metadata': {
                'dmark_ussd': {
                    'transaction_id':
                        self._request_defaults['transactionId'],
                    'transaction_time':
                        self._request_defaults['transactionTime'],
                    'creation_time':
                        self._request_defaults['creationTime'],
                },
            }
        }
        expected_field_values.update(field_values)

        for field, expected_value in expected_field_values.iteritems():
            self.assertEqual(msg[field], expected_value)

    def assert_ack(self, ack, reply):
        self.assertEqual(ack.payload['event_type'], 'ack')
        self.assertEqual(ack.payload['user_message_id'], reply['message_id'])
        self.assertEqual(ack.payload['sent_message_id'], reply['message_id'])

    def assert_nack(self, nack, reply, reason):
        self.assertEqual(nack.payload['event_type'], 'nack')
        self.assertEqual(nack.payload['user_message_id'], reply['message_id'])
        self.assertEqual(nack.payload['nack_reason'], reason)

    @inlineCallbacks
    def test_inbound_begin(self):
        user_content = "Who are you?"
        d = self.tx_helper.mk_request(ussdRequestString=user_content)
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.assert_inbound_message(
            msg,
            session_event=TransportUserMessage.SESSION_NEW,
            content=user_content)

        reply_content = "We are the Knights Who Say ... Ni!"
        reply = msg.reply(reply_content)
        self.tx_helper.dispatch_outbound(reply)
        response = yield d
        self.assertEqual(json.loads(response.delivered_body), {
            "responseString": reply_content,
            "action": "request",
        })
        self.assertEqual(response.code, 200)

        [ack] = yield self.tx_helper.wait_for_dispatched_events(1)
        self.assert_ack(ack, reply)

    @inlineCallbacks
    def test_inbound_status(self):
        d = self.tx_helper.mk_request()
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        [status] = yield self.tx_helper.get_dispatched_statuses()
        self.tx_helper.dispatch_outbound(msg.reply('foo'))
        yield d

        self.assertEqual(status['status'], 'ok')
        self.assertEqual(status['component'], 'request')
        self.assertEqual(status['type'], 'request_parsed')
        self.assertEqual(status['message'], 'Request parsed')

    @inlineCallbacks
    def test_inbound_cannot_decode(self):
        '''If the content cannot be decoded, an error shoould be sent back'''
        user_content = "Who are you?".encode('utf-32')
        response = yield self.tx_helper.mk_request(
            ussdRequestString=user_content)
        self.assertEqual(response.code, 400)

        body = json.loads(response.delivered_body)
        request = body['invalid_request']
        self.assertEqual(request['content'], '')
        self.assertEqual(request['path'], self.config['web_path'])
        self.assertEqual(request['method'], 'GET')
        self.assertEqual(request['headers']['Connection'], ['close'])
        encoded_str = urllib.urlencode({'ussdRequestString': user_content})
        self.assertTrue(encoded_str in request['uri'])

    @inlineCallbacks
    def test_inbound_cannot_decode_status(self):
        '''If the request cannot be decoded, a status event should be sent'''
        user_content = "Who are you?".encode('utf-32')
        yield self.tx_helper.mk_request(ussdRequestString=user_content)

        [status] = self.tx_helper.get_dispatched_statuses()
        self.assertEqual(status['component'], 'request')
        self.assertEqual(status['status'], 'down')
        self.assertEqual(status['type'], 'invalid_encoding')
        self.assertEqual(status['message'], 'Invalid encoding')

        request = status['details']['request']
        self.assertEqual(request['content'], '')
        self.assertEqual(request['path'], self.config['web_path'])
        self.assertEqual(request['method'], 'GET')
        self.assertEqual(request['headers']['Connection'], ['close'])
        encoded_str = urllib.urlencode({'ussdRequestString': user_content})
        self.assertTrue(encoded_str in request['uri'])

    @inlineCallbacks
    def test_inbound_resume_and_reply_with_end(self):
        yield self.mk_session(self._transaction_id)

        user_content = "Well, what is it you want?"
        d = self.tx_helper.mk_request(ussdRequestString=user_content)
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.assert_inbound_message(
            msg,
            session_event=TransportUserMessage.SESSION_RESUME,
            content=user_content)

        reply_content = "We want ... a shrubbery!"
        reply = msg.reply(reply_content, continue_session=False)
        self.tx_helper.dispatch_outbound(reply)
        response = yield d
        self.assertEqual(json.loads(response.delivered_body), {
            "responseString": reply_content,
            "action": "end",
        })
        self.assertEqual(response.code, 200)

        [ack] = yield self.tx_helper.wait_for_dispatched_events(1)
        self.assert_ack(ack, reply)

    @inlineCallbacks
    def test_inbound_resume_and_reply_with_resume(self):
        yield self.mk_session()

        user_content = "Well, what is it you want?"
        d = self.tx_helper.mk_request(ussdRequestString=user_content)
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.assert_inbound_message(
            msg,
            session_event=TransportUserMessage.SESSION_RESUME,
            content=user_content)

        reply_content = "We want ... a shrubbery!"
        reply = msg.reply(reply_content, continue_session=True)
        self.tx_helper.dispatch_outbound(reply)
        response = yield d
        self.assertEqual(json.loads(response.delivered_body), {
            "responseString": reply_content,
            "action": "request",
        })
        self.assertEqual(response.code, 200)

        [ack] = yield self.tx_helper.wait_for_dispatched_events(1)
        self.assert_ack(ack, reply)

    @inlineCallbacks
    def test_request_with_missing_parameters(self):
        response = yield self.tx_helper.mk_request_raw(
            params={"ussdServiceCode": '', "msisdn": '', "creationTime": ''})

        json_resp = json.loads(response.delivered_body)
        json_resp['missing_parameter'] = sorted(json_resp['missing_parameter'])
        self.assertEqual(json_resp, {
            'missing_parameter': sorted([
                "transactionTime", "transactionId", "response",
                "ussdRequestString",
            ]),
        })
        self.assertEqual(response.code, 400)

    @inlineCallbacks
    def test_status_with_missing_parameters(self):
        '''A request with missing parameters should send a TransportStatus
        with the relevant details.'''
        yield self.tx_helper.mk_request_raw(
            params={"ussdServiceCode": '', "msisdn": '', "creationTime": ''})

        [status] = yield self.tx_helper.get_dispatched_statuses()
        self.assertEqual(status['status'], 'down')
        self.assertEqual(status['component'], 'request')
        self.assertEqual(status['type'], 'invalid_inbound_fields')
        self.assertEqual(sorted(status['details']['missing_parameter']), [
            'response', 'transactionId', 'transactionTime',
            'ussdRequestString'])

    @inlineCallbacks
    def test_request_with_unexpected_parameters(self):
        response = yield self.tx_helper.mk_request(
            unexpected_p1='', unexpected_p2='')

        self.assertEqual(response.code, 400)
        body = json.loads(response.delivered_body)
        self.assertEqual(set(['unexpected_parameter']), set(body.keys()))
        self.assertEqual(
            sorted(body['unexpected_parameter']),
            ['unexpected_p1', 'unexpected_p2'])

    @inlineCallbacks
    def test_status_with_unexpected_parameters(self):
        '''A request with unexpected parameters should send a TransportStatus
        with the relevant details.'''
        yield self.tx_helper.mk_request(
            unexpected_p1='', unexpected_p2='')

        [status] = yield self.tx_helper.get_dispatched_statuses()
        self.assertEqual(status['status'], 'down')
        self.assertEqual(status['component'], 'request')
        self.assertEqual(status['type'], 'invalid_inbound_fields')
        self.assertEqual(sorted(status['details']['unexpected_parameter']), [
            'unexpected_p1', 'unexpected_p2'])

    @inlineCallbacks
    def test_nack_insufficient_message_fields(self):
        reply = self.tx_helper.make_outbound(
            None, message_id='23', in_reply_to=None)
        self.tx_helper.dispatch_outbound(reply)
        [nack] = yield self.tx_helper.wait_for_dispatched_events(1)
        self.assert_nack(nack, reply, 'Missing fields: in_reply_to, content')

    @inlineCallbacks
    def test_nack_http_http_response_failure(self):
        reply = self.tx_helper.make_outbound(
            'There are some who call me ... Tim!', message_id='23',
            in_reply_to='some-number')
        self.tx_helper.dispatch_outbound(reply)
        [nack] = yield self.tx_helper.wait_for_dispatched_events(1)
        self.assert_nack(
            nack, reply, 'Could not find original request.')

    @inlineCallbacks
    def test_status_quick_response(self):
        '''Ok status event should be sent if the response is quick.'''
        d = self.tx_helper.mk_request()
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        yield self.tx_helper.clear_dispatched_statuses()

        self.tx_helper.dispatch_outbound(msg.reply('foo'))
        yield d

        [status] = yield self.tx_helper.get_dispatched_statuses()
        self.assertEqual(status['status'], 'ok')
        self.assertEqual(status['component'], 'response')
        self.assertEqual(status['message'], 'Response sent')
        self.assertEqual(status['type'], 'response_sent')

    @inlineCallbacks
    def test_status_degraded_slow_response(self):
        '''A degraded status event should be sent if the response took longer
        than 1 second.'''
        d = self.tx_helper.mk_request()
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        yield self.tx_helper.clear_dispatched_statuses()

        self.clock.advance(self.transport.response_time_degraded + 0.1)

        self.tx_helper.dispatch_outbound(msg.reply('foo'))
        yield d

        [status] = yield self.tx_helper.get_dispatched_statuses()
        self.assertEqual(status['status'], 'degraded')
        self.assertTrue(
            str(self.transport.response_time_degraded) in status['reasons'][0])
        self.assertEqual(status['component'], 'response')
        self.assertEqual(status['type'], 'slow_response')
        self.assertEqual(status['message'], 'Slow response')

    @inlineCallbacks
    def test_status_down_very_slow_response(self):
        '''A down status event should be sent if the response took longer
        than 10 seconds.'''
        d = self.tx_helper.mk_request()
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        yield self.tx_helper.clear_dispatched_statuses()

        self.clock.advance(self.transport.response_time_down + 0.1)

        self.tx_helper.dispatch_outbound(msg.reply('foo'))
        yield d

        [status] = yield self.tx_helper.get_dispatched_statuses()
        self.assertEqual(status['status'], 'down')
        self.assertTrue(
            str(self.transport.response_time_down) in status['reasons'][0])
        self.assertEqual(status['component'], 'response')
        self.assertEqual(status['type'], 'very_slow_response')
        self.assertEqual(status['message'], 'Very slow response')

    @inlineCallbacks
    def test_no_response_status_for_message_not_found(self):
        '''If we cannot find the starting timestamp for a message, no status
        message should be sent'''
        reply = self.tx_helper.make_outbound(
            'There are some who call me ... Tim!', message_id='23',
            in_reply_to='some-number')
        self.tx_helper.dispatch_outbound(reply)
        statuses = yield self.tx_helper.get_dispatched_statuses()
        self.assertEqual(len(statuses), 0)

    @inlineCallbacks
    def test_no_good_status_event_for_bad_responses(self):
        '''If the http response is not a good (200-399) response, then a
        status event shouldn't be sent, because we send different status
        events for those errors.'''
        d = self.tx_helper.mk_request()

        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        yield self.tx_helper.clear_dispatched_statuses()

        self.transport.finish_request(msg['message_id'], '', code=500)

        yield d

        statuses = yield self.tx_helper.get_dispatched_statuses()
        self.assertEqual(len(statuses), 0)

    @inlineCallbacks
    def test_no_degraded_status_event_for_bad_responses(self):
        '''If the http response is not a good (200-399) response, then a
        status event shouldn't be sent, because we send different status
        events for those errors.'''
        d = self.tx_helper.mk_request()

        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        yield self.tx_helper.clear_dispatched_statuses()

        self.clock.advance(self.transport.response_time_degraded + 0.1)

        self.transport.finish_request(msg['message_id'], '', code=500)

        yield d

        statuses = yield self.tx_helper.get_dispatched_statuses()
        self.assertEqual(len(statuses), 0)

    @inlineCallbacks
    def test_no_down_status_event_for_bad_responses(self):
        '''If the http response is not a good (200-399) response, then a
        status event shouldn't be sent, because we send different status
        events for those errors.'''
        d = self.tx_helper.mk_request()

        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        yield self.tx_helper.clear_dispatched_statuses()

        self.clock.advance(self.transport.response_time_down + 0.1)

        self.transport.finish_request(msg['message_id'], '', code=500)

        yield d

        statuses = yield self.tx_helper.get_dispatched_statuses()
        self.assertEqual(len(statuses), 0)

    @inlineCallbacks
    def test_status_down_timeout(self):
        '''A down status event should be sent if the response timed out'''
        d = self.tx_helper.mk_request()
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        yield self.tx_helper.clear_dispatched_statuses()

        self.clock.advance(self.transport.request_timeout + 0.1)

        self.tx_helper.dispatch_outbound(msg.reply('foo'))
        yield d

        [status] = yield self.tx_helper.get_dispatched_statuses()
        self.assertEqual(status['status'], 'down')
        self.assertTrue(
            str(self.transport.request_timeout) in status['reasons'][0])
        self.assertEqual(status['component'], 'response')
        self.assertEqual(status['type'], 'timeout')
        self.assertEqual(status['message'], 'Response timed out')
        self.assertEqual(status['details'], {
            'response_time': self.transport.request_timeout + 0.1,
        })



# -*- test-case-name: vumi.demos.tests.test_words -*-

"""Demo ApplicationWorkers that perform simple text manipulations."""

from twisted.python import log

from vumi.application import ApplicationWorker


class SimpleAppWorker(ApplicationWorker):
    """Base class for very simple application workers.

       Configuration
       -------------
       transport_name : str
           Name of the transport.
       """

    def consume_user_message(self, msg):
        """Find or create a hangman game for this player.

        Then process the user's message.
        """
        content = msg['content'].encode('utf-8') if msg['content'] else None
        log.msg("User message: %s" % content)
        text = msg['content']
        if text is None:
            reply = self.get_help()
        else:
            reply = self.process_message(text)
        return self.reply_to(msg, reply)

    def process_message(self, text):
        raise NotImplementedError("Sub-classes should implement"
                                  " process_message.")

    def get_help(self):
        return "Enter text:"


class EchoWorker(SimpleAppWorker):
    """Echos text back to the sender."""

    def process_message(self, data):
        return data

    def get_help(self):
        return "Enter text to echo:"


class ReverseWorker(SimpleAppWorker):
    """Replies with reversed text."""

    def process_message(self, data):
        return data[::-1]

    def get_help(self):
        return "Enter text to reverse:"


class WordCountWorker(SimpleAppWorker):
    """Returns word and letter counts to the sender."""

    def process_message(self, data):
        response = []
        words = len(data.split())
        response.append("%s word%s" % (words, "s" * (words != 1)))
        chars = len(data)
        response.append("%s char%s" % (chars, "s" * (chars != 1)))
        return ', '.join(response)

    def get_help(self):
        return "Enter text to return word and character counts for:"



from twisted.internet.defer import inlineCallbacks

from vumi.demos.tictactoe import TicTacToeGame, TicTacToeWorker
from vumi.message import TransportUserMessage
from vumi.application.tests.helpers import ApplicationHelper
from vumi.tests.helpers import VumiTestCase


class TestTicTacToeGame(VumiTestCase):
    def get_game(self, moves=()):
        game = TicTacToeGame('pX')
        game.set_player_O('pO')
        for sid, x, y in moves:
            game.move(sid, x, y)
        return game

    def test_game_init(self):
        game = TicTacToeGame('pX')
        self.assertEquals('pX', game.player_X)
        self.assertEquals(None, game.player_O)
        self.assertEquals([[' '] * 3] * 3, game.board)

        game.set_player_O('pO')
        self.assertEquals('pX', game.player_X)
        self.assertEquals('pO', game.player_O)
        self.assertEquals([[' '] * 3] * 3, game.board)

    def test_move(self):
        game = self.get_game()
        expected_board = [[' ', ' ', ' '] for _i in range(3)]

        self.assertEqual((True, 'pO'), game.move('pX', 0, 0))
        expected_board[0][0] = 'X'
        self.assertEqual(expected_board, game.board)

        self.assertEqual((True, 'pX'), game.move('pO', 1, 0))
        expected_board[0][1] = 'O'
        self.assertEqual(expected_board, game.board)

        self.assertEqual((False, 'pO'), game.move('pX', 0, 0))
        self.assertEqual(expected_board, game.board)

    def test_draw_board(self):
        game = self.get_game(moves=[('pX', 0, 0), ('pO', 1, 0), ('pX', 1, 2)])
        self.assertEqual("+---+---+---+\n"
                         "| X | O |   |\n"
                         "+---+---+---+\n"
                         "|   |   |   |\n"
                         "+---+---+---+\n"
                         "|   | X |   |\n"
                         "+---+---+---+",
                         game.draw_board())

    def test_check_draw(self):
        game = self.get_game()
        for y, x in [(y, x) for x in range(3) for y in range(3)]:
            self.assertEqual(False, game.check_draw())
            game.move('pX' if x + y % 2 == 0 else 'pO', x, y)
        self.assertEqual(True, game.check_draw())

    def test_check_win(self):
        game = self.get_game()
        for i, (y, x) in enumerate([(0, 0), (1, 0),
                                    (1, 2), (1, 1),
                                    (0, 2), (0, 1),
                                    (2, 2)]):
            self.assertEqual(False, game.check_win())
            game.move('pX' if i % 2 == 0 else 'pO', x, y)
        self.assertEqual('X', game.check_win())


class TestTicTacToeWorker(VumiTestCase):

    @inlineCallbacks
    def setUp(self):
        self.app_helper = self.add_helper(ApplicationHelper(TicTacToeWorker))
        self.worker = yield self.app_helper.get_application({})

    def dispatch_start_message(self, from_addr):
        return self.app_helper.make_dispatch_inbound(
            None, from_addr=from_addr,
            session_event=TransportUserMessage.SESSION_NEW)

    @inlineCallbacks
    def test_new_sessions(self):
        self.assertEquals({}, self.worker.games)
        self.assertEquals(None, self.worker.open_game)

        user1 = '+27831234567'
        user2 = '+27831234568'

        yield self.dispatch_start_message(user1)
        self.assertNotEquals(None, self.worker.open_game)
        game = self.worker.open_game
        self.assertEquals({user1: game}, self.worker.games)

        yield self.dispatch_start_message(user2)
        self.assertEquals(None, self.worker.open_game)
        self.assertEquals({user1: game, user2: game}, self.worker.games)

        [msg] = self.app_helper.get_dispatched_outbound()
        self.assertTrue(msg['content'].startswith('+---+---+---+'))

    @inlineCallbacks
    def test_moves(self):
        user1 = '+27831234567'
        user2 = '+27831234568'

        yield self.dispatch_start_message(user1)
        game = self.worker.open_game
        yield self.dispatch_start_message(user2)
        self.assertEquals(1, len(self.app_helper.get_dispatched_outbound()))

        yield self.app_helper.make_dispatch_inbound('1', from_addr=user1)
        self.assertEquals(2, len(self.app_helper.get_dispatched_outbound()))

        yield self.app_helper.make_dispatch_inbound('2', from_addr=user2)
        self.assertEquals(3, len(self.app_helper.get_dispatched_outbound()))

        self.assertEqual('X', game.board[0][0])
        self.assertEqual('O', game.board[0][1])

    @inlineCallbacks
    def test_full_game(self):
        user1 = '+27831234567'
        user2 = '+27831234568'
        yield self.dispatch_start_message(user1)
        game = self.worker.open_game
        yield self.dispatch_start_message(user2)

        for user, content in [
                (user1, '1'), (user2, '4'),
                (user1, '2'), (user2, '5'),
                (user1, '3')]:
            yield self.app_helper.make_dispatch_inbound(
                content, from_addr=user)

        self.assertEqual('X', game.check_win())
        [end1, end2] = self.app_helper.get_dispatched_outbound()[-2:]
        self.assertEqual(user1, end1["to_addr"])
        self.assertEqual("You won!", end1["content"])
        self.assertEqual(user2, end2["to_addr"])
        self.assertEqual("You lost!", end2["content"])



# -*- test-case-name: vumi.tests.test_log -*-
import logging
from functools import partial

from twisted.python import log


debug = partial(log.msg, logLevel=logging.DEBUG)
info = partial(log.msg, logLevel=logging.INFO)
warning = partial(log.msg, logLevel=logging.WARNING)
error = partial(log.err, logLevel=logging.ERROR)
critical = partial(log.err, logLevel=logging.CRITICAL)

# make transition from twisted.python.log easier
msg = info
err = error


class WrappingLogger(object):
    '''A logger that will add the additional arguments that it is initialized
    with to every logging call.'''
    def __init__(self, **kwargs):
        self.debug = partial(debug, **kwargs)
        self.info = partial(info, **kwargs)
        self.warning = partial(warning, **kwargs)
        self.error = partial(error, **kwargs)
        self.critical = partial(critical, **kwargs)
        self.msg = partial(msg, **kwargs)
        self.err = partial(err, **kwargs)



# -*- test-case-name: vumi.transports.dmark.tests.test_dmark_ussd -*-

import json

from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.web import http

from vumi.components.session import SessionManager
from vumi.config import ConfigDict, ConfigInt
from vumi.message import TransportUserMessage
from vumi.transports.httprpc import HttpRpcTransport


class DmarkUssdTransportConfig(HttpRpcTransport.CONFIG_CLASS):
    """Config for Dmark USSD transport."""

    ussd_session_timeout = ConfigInt(
        "Number of seconds before USSD session information stored in Redis"
        " expires.",
        default=600, static=True)

    redis_manager = ConfigDict(
        "Redis client configuration.", default={}, static=True)


class DmarkUssdTransport(HttpRpcTransport):
    """Dmark USSD transport over HTTP.

    When a USSD message is received, Dmark will make an HTTP GET request to
    the transport with the following query parameters:

    * ``transactionId``: A unique ID for the USSD session (string).
    * ``msisdn``: The phone number that the message was sent from (string).
    * ``ussdServiceCode``: The USSD Service code the request was made to
      (string).
    * ``transactionTime``: The time the USSD request was received at Dmark,
      as a Unix timestamp (UTC).
    * ``ussdRequestString``: The full content of the USSD request(string).
    * ``creationTime``: The time the USSD request was sent, as a Unix
      timestamp (UTC), if available. (This time is given by the mobile
      network, and may not always be reliable.)
    * ``response``: ``"false"`` if this is a new session, ``"true"`` if it is
      not. Currently not used by the transport (it relies on the
      ``transactionId`` being unique instead).

    The transport may respond to this request either using JSON or form-encoded
    data. A successful response must return HTTP status code 200. Any other
    response code is treated as a failure.

    This transport responds with JSON encoded data. The JSON response
    contains the following keys:

    * ``responseString``: The content to be returned to the phone number that
      originated the USSD request.
    * ``action``: Either ``end`` or ``request``. ``end`` signifies that no
      further interaction is expected from the user and the USSD session should
      be closed. ``request`` signifies that further interaction is expected.

    **Example JSON response**:

    .. sourcecode: javascript

       {
         "responseString": "Hello from Vumi!",
         "action": "end"
       }
    """

    CONFIG_CLASS = DmarkUssdTransportConfig

    transport_type = 'ussd'

    ENCODING = 'utf-8'
    EXPECTED_FIELDS = frozenset([
        'transactionId', 'msisdn', 'ussdServiceCode', 'transactionTime',
        'ussdRequestString', 'creationTime', 'response',
    ])

    @inlineCallbacks
    def setup_transport(self):
        yield super(DmarkUssdTransport, self).setup_transport()
        config = self.get_static_config()
        r_prefix = "vumi.transports.dmark_ussd:%s" % self.transport_name
        self.session_manager = yield SessionManager.from_redis_config(
            config.redis_manager, r_prefix,
            max_session_length=config.ussd_session_timeout)

    @inlineCallbacks
    def teardown_transport(self):
        yield super(DmarkUssdTransport, self).teardown_transport()
        yield self.session_manager.stop()

    @inlineCallbacks
    def session_event_for_transaction(self, transaction_id):
        # XXX: There is currently no way to detect when the user closes
        #      the session (i.e. TransportUserMessage.SESSION_CLOSE)
        session_id = transaction_id
        session = yield self.session_manager.load_session(transaction_id)
        if session:
            session_event = TransportUserMessage.SESSION_RESUME
            yield self.session_manager.save_session(session_id, session)
        else:
            session_event = TransportUserMessage.SESSION_NEW
            yield self.session_manager.create_session(
                session_id, transaction_id=transaction_id)
        returnValue(session_event)

    @inlineCallbacks
    def handle_raw_inbound_message(self, request_id, request):
        try:
            values, errors = self.get_field_values(
                request, self.EXPECTED_FIELDS)
        except UnicodeDecodeError:
            self.log.msg('Bad request encoding: %r' % request)
            request_dict = {
                'uri': request.uri,
                'method': request.method,
                'path': request.path,
                'content': request.content.read(),
                'headers': dict(request.requestHeaders.getAllRawHeaders()),
            }
            self.finish_request(
                request_id, json.dumps({'invalid_request': request_dict}),
                code=http.BAD_REQUEST)
            yield self.add_status(
                component='request',
                status='down',
                type='invalid_encoding',
                message='Invalid encoding',
                details={
                    'request': request_dict,
                })
            return

        if errors:
            self.log.msg('Unhappy incoming message: %r' % (errors,))
            self.finish_request(
                request_id, json.dumps(errors), code=http.BAD_REQUEST)
            yield self.add_status(
                component='request',
                status='down',
                type='invalid_inbound_fields',
                message='Invalid inbound fields',
                details=errors)
            return

        yield self.add_status(
            component='request',
            status='ok',
            type='request_parsed',
            message='Request parsed',)

        to_addr = values["ussdServiceCode"]
        from_addr = values["msisdn"]
        session_event = yield self.session_event_for_transaction(
            values["transactionId"])

        yield self.publish_message(
            message_id=request_id,
            content=values["ussdRequestString"],
            to_addr=to_addr,
            from_addr=from_addr,
            provider='dmark',
            session_event=session_event,
            transport_type=self.transport_type,
            transport_metadata={
                'dmark_ussd': {
                    'transaction_id': values['transactionId'],
                    'transaction_time': values['transactionTime'],
                    'creation_time': values['creationTime'],
                }
            })

    @inlineCallbacks
    def handle_outbound_message(self, message):
        self.emit("DmarkUssdTransport consuming %r" % (message,))
        missing_fields = self.ensure_message_values(
            message, ['in_reply_to', 'content'])
        if missing_fields:
            nack = yield self.reject_message(message, missing_fields)
            returnValue(nack)

        if message["session_event"] == TransportUserMessage.SESSION_CLOSE:
            action = "end"
        else:
            action = "request"

        response_data = {
            "responseString": message["content"],
            "action": action,
        }

        response_id = self.finish_request(
            message['in_reply_to'], json.dumps(response_data))

        if response_id is not None:
            ack = yield self.publish_ack(
                user_message_id=message['message_id'],
                sent_message_id=message['message_id'])
            returnValue(ack)
        else:
            nack = yield self.publish_nack(
                user_message_id=message['message_id'],
                sent_message_id=message['message_id'],
                reason="Could not find original request.")
            returnValue(nack)

    def on_down_response_time(self, message_id, time):
        request = self.get_request(message_id)
        # We send different status events for error responses
        if request.code < 200 or request.code >= 300:
            return
        return self.add_status(
            component='response',
            status='down',
            type='very_slow_response',
            message='Very slow response',
            reasons=[
                'Response took longer than %fs' % (
                    self.response_time_down,)
            ],
            details={
                'response_time': time,
            })

    def on_degraded_response_time(self, message_id, time):
        request = self.get_request(message_id)
        # We send different status events for error responses
        if request.code < 200 or request.code >= 300:
            return
        return self.add_status(
            component='response',
            status='degraded',
            type='slow_response',
            message='Slow response',
            reasons=[
                'Response took longer than %fs' % (
                    self.response_time_degraded,)
            ],
            details={
                'response_time': time,
            })

    def on_good_response_time(self, message_id, time):
        request = self.get_request(message_id)
        # We send different status events for error responses
        if request.code < 200 or request.code >= 400:
            return
        return self.add_status(
            component='response',
            status='ok',
            type='response_sent',
            message='Response sent',
            details={
                'response_time': time,
            })

    def on_timeout(self, message_id, time):
        return self.add_status(
            component='response',
            status='down',
            type='timeout',
            message='Response timed out',
            reasons=[
                'Response took longer than %fs' % (
                    self.request_timeout,)
            ],
            details={
                'response_time': time,
            })



# -*- coding: utf-8 -*-

"""Tests for vumi.persist.fields."""

from datetime import datetime
from functools import wraps

from twisted.internet.defer import inlineCallbacks, returnValue

from vumi.message import Message, TransportUserMessage
from vumi.persist.fields import (
    ValidationError, Field, Integer, Unicode, Tag, Timestamp, Json,
    ListOf, SetOf, Dynamic, FieldWithSubtype, Boolean, VumiMessage,
    ForeignKey, ManyToMany, ComputedValue)
from vumi.persist.model import Manager, Model
from vumi.tests.helpers import VumiTestCase, MessageHelper, import_skip


def needs_riak(method):
    """
    Mark a test method as needing Riak setup.
    """
    method.needs_riak = True
    return method


class ModelFieldTestsDecorator(object):
    """
    Class decorator for replacing `@needs_riak`-marked test methods with two
    wrapped versions, one for each Riak manager.

    This is used here instead of the more usual mechanism of a mixin and two
    subclasses because we have a lot of small test classes which have several
    test methods that don't need Riak.
    """

    def __call__(deco, cls):
        """
        Find all methods on the the given class that are marked with
        `@needs_riak` and replace them with wrapped versions for both
        RiakManager and TxRiakManager.
        """
        # We can't use `inspect.getmembers()` because of a bug in Python 2.6
        # around empty slots: http://bugs.python.org/issue1162154
        needs_riak_methods = []
        for member_name in dir(cls):
            # If the class has an empty slot (`__provides__` from
            # zope.interface, in this case) we get a name for a member that
            # does not exist.
            member = getattr(cls, member_name, None)
            if getattr(member, "needs_riak", False):
                needs_riak_methods.append((member_name, member))
        for name, meth in needs_riak_methods:
            delattr(cls, name)
            setattr(cls, name + "__on_riak", deco.wrap_riak_setup(meth))
            setattr(cls, name + "__on_txriak", deco.wrap_txriak_setup(meth))
        return cls

    def wrap_riak_setup(deco, meth):
        """
        Return a wrapper around `meth` that sets up a RiakManager.
        """
        @wraps(meth)
        def wrapper(self):
            deco.setup_riak(self)
            return meth(self)
        return wrapper

    def wrap_txriak_setup(deco, meth):
        """
        Return a wrapper around `meth` that sets up a TxRiakManager.
        """
        @wraps(meth)
        def wrapper(self):
            d = deco.setup_txriak(self)
            return d.addCallback(lambda _: meth(self))
        return wrapper

    def setup_riak(deco, self):
        """
        Set up a RiakManager on the given test class.
        """
        try:
            from vumi.persist.riak_manager import RiakManager
        except ImportError, e:
            import_skip(e, 'riak')
        self.manager = RiakManager.from_config({'bucket_prefix': 'test.'})
        self.add_cleanup(deco.cleanup_manager, self)
        self.manager.purge_all()

    @inlineCallbacks
    def setup_txriak(deco, self):
        """
        Set up a TxRiakManager on the given test class.
        """
        try:
            from vumi.persist.txriak_manager import TxRiakManager
        except ImportError, e:
            import_skip(e, 'riak')
        self.manager = TxRiakManager.from_config({'bucket_prefix': 'test.'})
        self.add_cleanup(deco.cleanup_manager, self)
        yield self.manager.purge_all()

    @inlineCallbacks
    def cleanup_manager(deco, self):
        """
        Clean up the Riak manager on the given test class.
        """
        yield self.manager.purge_all()
        yield self.manager.close_manager()


model_field_tests = ModelFieldTestsDecorator()


def watch_model_changes(modelobj):
    changed_field_names = []
    old_field_changed = modelobj._field_changed

    def new_field_changed(changed_field_name):
        changed_field_names.append(changed_field_name)
        return old_field_changed

    modelobj._field_changed = new_field_changed
    return changed_field_names


@model_field_tests
class TestBaseField(VumiTestCase):
    def test_validate(self):
        f = Field()
        f.validate("foo")
        f.validate(object())
        self.assertRaises(ValidationError, f.validate, None)

    def test_validate_null(self):
        f = Field(null=True)
        f.validate("foo")
        f.validate(None)

    def test_to_riak(self):
        f = Field()
        obj = object()
        self.assertEqual(f.to_riak(obj), obj)

    def test_from_riak(self):
        f = Field()
        obj = object()
        self.assertEqual(f.from_riak(obj), obj)

    def test_get_descriptor(self):
        f = Field()
        descriptor = f.get_descriptor("foo")
        self.assertEqual(descriptor.key, "foo")
        self.assertEqual(descriptor.field, f)
        self.assertTrue("Field object" in repr(descriptor))

    class BaseFieldModel(Model):
        """
        Toy model for Field tests.
        """
        f = Field()

    @needs_riak
    @Manager.calls_manager
    def test_assorted_values(self):
        """
        Values are preserved when the field is stored and later loaded.
        """
        base_model = self.manager.proxy(self.BaseFieldModel)
        yield base_model("m_str", f="string").save()
        yield base_model("m_int", f=1).save()
        yield base_model("m_list", f=["string", 1]).save()
        yield base_model("m_dict", f={"key": "val"}).save()

        m_str = yield base_model.load("m_str")
        self.assertEqual(m_str.f, "string")
        m_int = yield base_model.load("m_int")
        self.assertEqual(m_int.f, 1)
        m_list = yield base_model.load("m_list")
        self.assertEqual(m_list.f, ["string", 1])
        m_dict = yield base_model.load("m_dict")
        self.assertEqual(m_dict.f, {"key": "val"})

    @needs_riak
    @Manager.calls_manager
    def test_field_changed_notification(self):
        """
        The model object is notified about changes to the field value.
        """
        base_model = self.manager.proxy(self.BaseFieldModel)
        m_str = base_model("m_str", f="string")
        field_changes = watch_model_changes(m_str)

        self.assertEqual(field_changes, [])
        # Field set to previous value, no notification.
        m_str.f = "string"
        self.assertEqual(field_changes, [])
        # Field set to new value, notification sent.
        m_str.f = "string2"
        self.assertEqual(field_changes, ["f"])
        # Model saved, no notification.
        yield m_str.save()
        self.assertEqual(field_changes, ["f"])


@model_field_tests
class TestInteger(VumiTestCase):
    def test_validate_unbounded(self):
        i = Integer()
        i.validate(5)
        i.validate(-3)
        self.assertRaises(ValidationError, i.validate, 5.0)
        self.assertRaises(ValidationError, i.validate, "5")

    def test_validate_minimum(self):
        i = Integer(min=3)
        i.validate(3)
        i.validate(4)
        self.assertRaises(ValidationError, i.validate, 2)

    def test_validate_maximum(self):
        i = Integer(max=5)
        i.validate(5)
        i.validate(4)
        self.assertRaises(ValidationError, i.validate, 6)

    class IntegerModel(Model):
        """
        Toy model for Integer field tests.
        """
        i = Integer()

    @needs_riak
    @Manager.calls_manager
    def test_assorted_values(self):
        """
        Values are preserved when the field is stored and later loaded.
        """
        int_model = self.manager.proxy(self.IntegerModel)
        yield int_model("m_1", i=1).save()
        yield int_model("m_leet", i=1337).save()

        m_1 = yield int_model.load("m_1")
        self.assertEqual(m_1.i, 1)
        m_leet = yield int_model.load("m_leet")
        self.assertEqual(m_leet.i, 1337)


@model_field_tests
class TestBoolean(VumiTestCase):

    def test_validate(self):
        b = Boolean()
        b.validate(True)
        b.validate(False)
        self.assertRaises(ValidationError, b.validate, 'True')
        self.assertRaises(ValidationError, b.validate, 'False')
        self.assertRaises(ValidationError, b.validate, 1)
        self.assertRaises(ValidationError, b.validate, 0)

    class BooleanModel(Model):
        """
        Toy model for Boolean field tests.
        """
        b = Boolean()

    @needs_riak
    @Manager.calls_manager
    def test_assorted_values(self):
        """
        Values are preserved when the field is stored and later loaded.
        """
        bool_model = self.manager.proxy(self.BooleanModel)
        yield bool_model("m_t", b=True).save()
        yield bool_model("m_f", b=False).save()

        m_t = yield bool_model.load("m_t")
        self.assertEqual(m_t.b, True)
        m_f = yield bool_model.load("m_f")
        self.assertEqual(m_f.b, False)


@model_field_tests
class TestUnicode(VumiTestCase):
    def test_validate(self):
        u = Unicode()
        u.validate(u"")
        u.validate(u"a")
        u.validate(u"æ")
        u.validate(u"foé")
        self.assertRaises(ValidationError, u.validate, "")
        self.assertRaises(ValidationError, u.validate, "foo")
        self.assertRaises(ValidationError, u.validate, 3)

    def test_validate_max_length(self):
        u = Unicode(max_length=5)
        u.validate(u"12345")
        u.validate(u"1234")
        self.assertRaises(ValidationError, u.validate, u"123456")

    class UnicodeModel(Model):
        """
        Toy model for Unicode field tests.
        """
        u = Unicode()

    @needs_riak
    @Manager.calls_manager
    def test_assorted_values(self):
        """
        Values are preserved when the field is stored and later loaded.
        """
        unicode_model = self.manager.proxy(self.UnicodeModel)
        yield unicode_model("m_empty", u=u"").save()
        yield unicode_model("m_full", u=u"You must be an optimist").save()
        yield unicode_model("m_unicode", u=u"foé").save()

        m_empty = yield unicode_model.load("m_empty")
        self.assertEqual(m_empty.u, u"")
        m_full = yield unicode_model.load("m_full")
        self.assertEqual(m_full.u, u"You must be an optimist")
        m_unicode = yield unicode_model.load("m_unicode")
        self.assertEqual(m_unicode.u, u"foé")


@model_field_tests
class TestTag(VumiTestCase):
    def test_validate(self):
        t = Tag()
        t.validate(("pool", "tagname"))
        self.assertRaises(ValidationError, t.validate, ["pool", "tagname"])
        self.assertRaises(ValidationError, t.validate, ("pool",))

    def test_to_riak(self):
        t = Tag()
        self.assertEqual(t.to_riak(("pool", "tagname")), ["pool", "tagname"])

    def test_from_riak(self):
        t = Tag()
        self.assertEqual(t.from_riak(["pool", "tagname"]), ("pool", "tagname"))


@model_field_tests
class TestTimestamp(VumiTestCase):
    def test_validate(self):
        t = Timestamp()
        t.validate(datetime.now())
        t.validate("2007-01-25T12:00:00Z")
        t.validate(u"2007-01-25T12:00:00Z")
        self.assertRaises(ValidationError, t.validate, "foo")

    def test_to_riak(self):
        t = Timestamp()
        dt = datetime(2100, 10, 5, 11, 10, 9)
        self.assertEqual(t.to_riak(dt), "2100-10-05 11:10:09.000000")

    def test_from_riak(self):
        t = Timestamp()
        dt = datetime(2100, 10, 5, 11, 10, 9)
        self.assertEqual(t.from_riak("2100-10-05 11:10:09.000000"), dt)

    class TimestampModel(Model):
        """
        Toy model for Timestamp tests.
        """
        time = Timestamp(null=True)

    @needs_riak
    def test_set_field(self):
        """
        A timestamp field can be set to a datetime or string value through its
        descriptor.
        """
        timestamp_model = self.manager.proxy(self.TimestampModel)
        t = timestamp_model("foo")

        now = datetime.now()
        t.time = now
        self.assertEqual(t.time, now)

        t.time = u"2007-01-25T12:00:00Z"
        self.assertEqual(t.time, datetime(2007, 01, 25, 12, 0))

    @needs_riak
    @Manager.calls_manager
    def test_assorted_values(self):
        """
        Values are preserved when the field is stored and later loaded.
        """
        timestamp_model = self.manager.proxy(self.TimestampModel)
        now = datetime.now()
        yield timestamp_model("m_now", time=now).save()
        yield timestamp_model("m_string", time=u"2007-01-25T12:00:00Z").save()

        m_now = yield timestamp_model.load("m_now")
        self.assertEqual(m_now.time, now)
        m_string = yield timestamp_model.load("m_string")
        self.assertEqual(m_string.time, datetime(2007, 01, 25, 12, 0))


@model_field_tests
class TestJson(VumiTestCase):
    def test_validate(self):
        j = Json()
        j.validate({"foo": None})
        self.assertRaises(ValidationError, j.validate, None)

    def test_to_riak(self):
        j = Json()
        d = {"foo": 5}
        self.assertEqual(j.to_riak(d), d)

    def test_from_riak(self):
        j = Json()
        d = {"foo": [1, 2, 3]}
        self.assertEqual(j.from_riak(d), d)

    class JsonModel(Model):
        """
        Toy model for Json tests.
        """
        j = Json()

    @needs_riak
    @Manager.calls_manager
    def test_assorted_values(self):
        """
        Values are preserved when the field is stored and later loaded.
        """
        json_model = self.manager.proxy(self.JsonModel)
        yield json_model("m_str", j="string").save()
        yield json_model("m_int", j=1).save()
        yield json_model("m_list", j=["string", 1]).save()
        yield json_model("m_dict", j={"key": "val"}).save()

        m_str = yield json_model.load("m_str")
        self.assertEqual(m_str.j, "string")
        m_int = yield json_model.load("m_int")
        self.assertEqual(m_int.j, 1)
        m_list = yield json_model.load("m_list")
        self.assertEqual(m_list.j, ["string", 1])
        m_dict = yield json_model.load("m_dict")
        self.assertEqual(m_dict.j, {"key": "val"})


class TestFieldWithSubtype(VumiTestCase):
    def test_fails_on_fancy_subtype(self):
        self.assertRaises(RuntimeError, FieldWithSubtype, Dynamic())


@model_field_tests
class TestDynamic(VumiTestCase):
    def test_validate(self):
        dynamic = Dynamic()
        dynamic.validate({u'a': u'foo', u'b': u'bar'})
        self.assertRaises(ValidationError, dynamic.validate,
                          {u'a': 'foo', u'b': u'bar'})
        self.assertRaises(ValidationError, dynamic.validate,
                          u'this is not a dict')
        self.assertRaises(ValidationError, dynamic.validate,
                          {u'a': 'foo', u'b': 2})

    class DynamicModel(Model):
        """
        Toy model for Dynamic tests.
        """
        a = Unicode()
        contact_info = Dynamic()

    @needs_riak
    def test_get_data_with_dynamic_proxy(self):
        """
        A Dynamic field creates more than one field in the Riak object.
        """
        dynamic_model = self.manager.proxy(self.DynamicModel)

        m = dynamic_model("foo", a=u"ab")
        m.contact_info['foo'] = u'bar'
        m.contact_info['zip'] = u'zap'

        self.assertEqual(m.get_data(), {
            'key': 'foo',
            '$VERSION': None,
            'a': 'ab',
            'contact_info.foo': 'bar',
            'contact_info.zip': 'zap',
        })

    def _create_dynamic_instance(self, dynamic_model, key):
        m = dynamic_model(key, a=u"ab")
        m.contact_info['cellphone'] = u"+27123"
        m.contact_info['telephone'] = u"+2755"
        m.contact_info['honorific'] = u"BDFL"
        return m

    @needs_riak
    @Manager.calls_manager
    def test_dynamic_value(self):
        """
        Values are preserved when the field is stored and later loaded.
        """
        dynamic_model = self.manager.proxy(self.DynamicModel)
        yield self._create_dynamic_instance(dynamic_model, "foo").save()

        m = yield dynamic_model.load("foo")
        self.assertEqual(m.a, u"ab")
        self.assertEqual(m.contact_info['cellphone'], u"+27123")
        self.assertEqual(m.contact_info['telephone'], u"+2755")
        self.assertEqual(m.contact_info['honorific'], u"BDFL")

    @needs_riak
    def test_dynamic_field_init(self):
        """
        Dynamic fields can be initialised with dicts.
        """
        dynamic_model = self.manager.proxy(self.DynamicModel)
        contact_info = {'cellphone': u'+27123', 'telephone': u'+2755'}
        m = dynamic_model("foo", a=u"ab", contact_info=contact_info)
        self.assertEqual(m.contact_info.copy(), contact_info)

    @needs_riak
    def test_dynamic_field_keys_and_values(self):
        """
        Dynamic field keys and values are available as lists or and iterators,
        similar to a dict.
        """
        dynamic_model = self.manager.proxy(self.DynamicModel)
        m = self._create_dynamic_instance(dynamic_model, "foo")

        keys = m.contact_info.keys()
        iterkeys = m.contact_info.iterkeys()
        self.assertTrue(isinstance(keys, list))
        self.assertTrue(hasattr(iterkeys, 'next'))
        self.assertEqual(sorted(keys), ['cellphone', 'honorific', 'telephone'])
        self.assertEqual(sorted(iterkeys), sorted(keys))

        values = m.contact_info.values()
        itervalues = m.contact_info.itervalues()
        self.assertTrue(isinstance(values, list))
        self.assertTrue(hasattr(itervalues, 'next'))
        self.assertEqual(sorted(values), ["+27123", "+2755", "BDFL"])
        self.assertEqual(sorted(itervalues), sorted(values))

        items = m.contact_info.items()
        iteritems = m.contact_info.iteritems()
        self.assertTrue(isinstance(items, list))
        self.assertTrue(hasattr(iteritems, 'next'))
        self.assertEqual(sorted(items), [('cellphone', "+27123"),
                                         ('honorific', "BDFL"),
                                         ('telephone', "+2755")])
        self.assertEqual(sorted(iteritems), sorted(items))

    @needs_riak
    def test_dynamic_field_clear(self):
        """
        Dynamic fields can be cleared.
        """
        dynamic_model = self.manager.proxy(self.DynamicModel)
        m = self._create_dynamic_instance(dynamic_model, "foo")
        m.contact_info.clear()
        self.assertEqual(m.contact_info.items(), [])

    @needs_riak
    def test_dynamic_field_update(self):
        """
        Dynamic fields can be bulk-updated.
        """
        dynamic_model = self.manager.proxy(self.DynamicModel)
        m = self._create_dynamic_instance(dynamic_model, "foo")
        m.contact_info.update({"cellphone": "123", "name": "foo"})
        self.assertEqual(sorted(m.contact_info.items()), [
            ('cellphone', "123"), ('honorific', "BDFL"), ('name', "foo"),
            ('telephone', "+2755")])

    @needs_riak
    def test_dynamic_field_contains(self):
        """
        Dynamic fields support `in`.
        """
        dynamic_model = self.manager.proxy(self.DynamicModel)
        m = self._create_dynamic_instance(dynamic_model, "foo")
        self.assertTrue("cellphone" in m.contact_info)
        self.assertFalse("landline" in m.contact_info)

    @needs_riak
    def test_dynamic_field_del(self):
        """
        Values can be removed from dynamic fields.
        """
        dynamic_model = self.manager.proxy(self.DynamicModel)
        m = self._create_dynamic_instance(dynamic_model, "foo")
        del m.contact_info["telephone"]
        self.assertEqual(sorted(m.contact_info.keys()),
                         ['cellphone', 'honorific'])

    @needs_riak
    def test_dynamic_field_setting(self):
        """
        Setting a dynamic field to a dict replaces its contents.
        """
        dynamic_model = self.manager.proxy(self.DynamicModel)
        m = self._create_dynamic_instance(dynamic_model, "foo")
        m.contact_info = {u'cellphone': u'789', u'name': u'foo'}
        self.assertEqual(sorted(m.contact_info.items()), [
            (u'cellphone', u'789'),
            (u'name', u'foo'),
        ])


@model_field_tests
class TestListOf(VumiTestCase):
    def test_validate(self):
        """
        By default, a ListOf field is a list of Unicode fields.
        """
        listof = ListOf()
        listof.validate([u'foo', u'bar'])
        self.assertRaises(ValidationError, listof.validate,
                          u'this is not a list')
        self.assertRaises(ValidationError, listof.validate, ['a', 2])
        self.assertRaises(ValidationError, listof.validate, [1, 2])

    def test_validate_with_subtype(self):
        """
        If an explicit subtype is provided, its validation is used.
        """
        listof_unicode = ListOf(Unicode())
        listof_unicode.validate([u"a", u"b"])
        self.assertRaises(ValidationError, listof_unicode.validate, [1, 2])

        listof_int = ListOf(Integer())
        listof_int.validate([1, 2])
        self.assertRaises(ValidationError, listof_int.validate, [u"a", u"b"])

        listof_smallint = ListOf(Integer(max=10))
        listof_smallint.validate([1, 2])
        self.assertRaises(
            ValidationError, listof_smallint.validate, [1, 100])

    class ListOfModel(Model):
        """
        Toy model for ListOf tests.
        """
        items = ListOf(Integer())
        texts = ListOf(Unicode())

    class IndexedListOfModel(Model):
        """
        Toy model for ListOf index tests.
        """
        items = ListOf(Integer(), index=True)

    @needs_riak
    def test_get_data_with_list_proxy(self):
        """
        A ListOf field creates a list field in the Riak object.
        """
        list_model = self.manager.proxy(self.ListOfModel)
        m = list_model("foo")
        m.items.append(1)
        m.items.append(42)
        m.texts.append(u"Thing 1.")
        m.texts.append(u"Thing 42.")
        self.assertEqual(m.get_data(), {
            'key': 'foo',
            '$VERSION': None,
            'items': [1, 42],
            'texts': [u"Thing 1.", u"Thing 42."],
        })

    @needs_riak
    @Manager.calls_manager
    def test_listof_fields(self):
        """
        A ListOf field can be manipulated as if it were a list.
        """
        list_model = self.manager.proxy(self.ListOfModel)
        l1 = list_model("foo")
        l1.items.append(1)
        l1.items.append(2)
        yield l1.save()

        l2 = yield list_model.load("foo")
        self.assertEqual(l2.items[0], 1)
        self.assertEqual(l2.items[1], 2)
        self.assertEqual(list(l2.items), [1, 2])

        l2.items[0] = 5
        self.assertEqual(l2.items[0], 5)

        del l2.items[0]
        self.assertEqual(list(l2.items), [2])

        l2.items.append(5)
        self.assertEqual(list(l2.items), [2, 5])
        l2.items.remove(5)
        self.assertEqual(list(l2.items), [2])

        l2.items.extend([3, 4, 5])
        self.assertEqual(list(l2.items), [2, 3, 4, 5])

        l2.items = [1]
        self.assertEqual(list(l2.items), [1])

    @needs_riak
    @Manager.calls_manager
    def test_listof_fields_indexes(self):
        """
        An indexed ListOf field has an index value for each item in the list.
        """
        list_model = self.manager.proxy(self.IndexedListOfModel)
        l1 = list_model("foo")
        l1.items.append(1)
        l1.items.append(2)
        yield l1.save()

        assert_indexes = lambda mdl, values: self.assertEqual(
            mdl._riak_object.get_indexes(),
            set(('items_bin', str(v)) for v in values))

        l2 = yield list_model.load("foo")
        self.assertEqual(l2.items[0], 1)
        self.assertEqual(l2.items[1], 2)
        self.assertEqual(list(l2.items), [1, 2])
        assert_indexes(l2, [1, 2])

        l2.items[0] = 5
        self.assertEqual(l2.items[0], 5)
        assert_indexes(l2, [2, 5])

        del l2.items[0]
        self.assertEqual(list(l2.items), [2])
        assert_indexes(l2, [2])

        l2.items.append(5)
        self.assertEqual(list(l2.items), [2, 5])
        assert_indexes(l2, [2, 5])
        l2.items.remove(5)
        self.assertEqual(list(l2.items), [2])
        assert_indexes(l2, [2])

        l2.items.extend([3, 4, 5])
        self.assertEqual(list(l2.items), [2, 3, 4, 5])
        assert_indexes(l2, [2, 3, 4, 5])

        l2.items = [1]
        self.assertEqual(list(l2.items), [1])
        assert_indexes(l2, [1])


@model_field_tests
class TestSetOf(VumiTestCase):
    def test_validate(self):
        """
        By default, a SetOf field is a set of Unicode fields.
        """
        f = SetOf()
        f.validate(set([u'foo', u'bar']))
        self.assertRaises(ValidationError, f.validate, u'this is not a set')
        self.assertRaises(ValidationError, f.validate, set(['a', 2]))
        self.assertRaises(ValidationError, f.validate, [u'a', u'b'])

    def test_validate_with_subtype(self):
        """
        If an explicit subtype is provided, its validation is used.
        """
        setof_unicode = SetOf(Unicode())
        setof_unicode.validate(set([u"a", u"b"]))
        self.assertRaises(ValidationError, setof_unicode.validate, set([1, 2]))

        setof_int = SetOf(Integer())
        setof_int.validate(set([1, 2]))
        self.assertRaises(
            ValidationError, setof_int.validate, set([u"a", u"b"]))

        setof_smallint = SetOf(Integer(max=10))
        setof_smallint.validate(set([1, 2]))
        self.assertRaises(
            ValidationError, setof_smallint.validate, set([1, 100]))

    def test_to_riak(self):
        """
        The JSON representation of a SetOf field is a sorted list.
        """
        f = SetOf()
        self.assertEqual(f.to_riak(set([1, 2, 3])), [1, 2, 3])

    def test_from_riak(self):
        """
        The JSON list is turned into a set when read.
        """
        f = SetOf()
        self.assertEqual(f.from_riak([1, 2, 3]), set([1, 2, 3]))

    class SetOfModel(Model):
        """
        Toy model for SetOf tests.
        """
        items = SetOf(Integer())
        texts = SetOf(Unicode())

    class IndexedSetOfModel(Model):
        """
        Toy model for SetOf index tests.
        """
        items = SetOf(Integer(), index=True)

    @needs_riak
    def test_setof_fields_validation(self):
        set_model = self.manager.proxy(self.SetOfModel)
        m1 = set_model("foo")

        self.assertRaises(ValidationError, m1.items.add, "foo")
        self.assertRaises(ValidationError, m1.items.remove, "foo")
        self.assertRaises(ValidationError, m1.items.discard, "foo")
        self.assertRaises(ValidationError, m1.items.update, set(["foo"]))

    @needs_riak
    @Manager.calls_manager
    def test_setof_fields(self):
        """
        A SetOf field can be manipulated as if it were a set.
        """
        set_model = self.manager.proxy(self.SetOfModel)
        m1 = set_model("foo")
        m1.items.add(1)
        m1.items.add(2)
        yield m1.save()

        m2 = yield set_model.load("foo")
        self.assertTrue(1 in m2.items)
        self.assertTrue(2 in m2.items)
        self.assertEqual(set(m2.items), set([1, 2]))

        m2.items.add(5)
        self.assertTrue(5 in m2.items)

        m2.items.remove(1)
        self.assertTrue(1 not in m2.items)
        self.assertRaises(KeyError, m2.items.remove, 1)

        m2.items.add(1)
        m2.items.discard(1)
        self.assertTrue(1 not in m2.items)
        m2.items.discard(1)
        self.assertTrue(1 not in m2.items)

        m2.items.update([3, 4, 5])
        self.assertEqual(set(m2.items), set([2, 3, 4, 5]))

        m2.items = set([7, 8])
        self.assertEqual(set(m2.items), set([7, 8]))

    @needs_riak
    @Manager.calls_manager
    def test_setof_fields_indexes(self):
        """
        An indexed SetOf field has an index value for each item in the set.
        """
        set_model = self.manager.proxy(self.IndexedSetOfModel)
        m1 = set_model("foo")
        m1.items.add(1)
        m1.items.add(2)
        yield m1.save()

        assert_indexes = lambda mdl, values: self.assertEqual(
            mdl._riak_object.get_indexes(),
            set(('items_bin', str(v)) for v in values))

        m2 = yield set_model.load("foo")
        self.assertTrue(1 in m2.items)
        self.assertTrue(2 in m2.items)
        self.assertEqual(set(m2.items), set([1, 2]))
        assert_indexes(m2, [1, 2])

        m2.items.add(5)
        self.assertTrue(5 in m2.items)
        assert_indexes(m2, [1, 2, 5])

        m2.items.remove(1)
        self.assertTrue(1 not in m2.items)
        assert_indexes(m2, [2, 5])

        m2.items.add(1)
        m2.items.discard(1)
        self.assertTrue(1 not in m2.items)
        assert_indexes(m2, [2, 5])
        m2.items.discard(1)
        self.assertTrue(1 not in m2.items)
        assert_indexes(m2, [2, 5])

        m2.items.update([3, 4, 5])
        self.assertEqual(set(m2.items), set([2, 3, 4, 5]))
        assert_indexes(m2, [2, 3, 4, 5])

        m2.items = set([7, 8])
        self.assertEqual(set(m2.items), set([7, 8]))
        assert_indexes(m2, [7, 8])


@model_field_tests
class TestVumiMessage(VumiTestCase):
    def test_validate(self):
        f = VumiMessage(Message)
        msg = Message()
        f.validate(msg)
        self.assertRaises(
            ValidationError, f.validate, u'this is not a vumi message')
        self.assertRaises(
            ValidationError, f.validate, None)

    class VumiMessageModel(Model):
        """
        Toy model for VumiMessage tests.
        """
        msg = VumiMessage(TransportUserMessage)

    @needs_riak
    @Manager.calls_manager
    def test_vumimessage_field(self):
        msg_helper = self.add_helper(MessageHelper())
        msg_model = self.manager.proxy(self.VumiMessageModel)
        msg = msg_helper.make_inbound("foo", extra="bar")
        m1 = msg_model("foo", msg=msg)
        yield m1.save()

        m2 = yield msg_model.load("foo")
        self.assertEqual(m1.msg, m2.msg)
        self.assertEqual(m2.msg, msg)

        self.assertRaises(ValidationError, setattr, m1, "msg", "foo")

        # test extra keys are removed
        msg2 = msg_helper.make_inbound("foo")
        m1.msg = msg2
        self.assertTrue("extra" not in m1.msg)

    @needs_riak
    @Manager.calls_manager
    def test_vumimessage_field_excludes_cache(self):
        msg_helper = self.add_helper(MessageHelper())
        msg_model = self.manager.proxy(self.VumiMessageModel)
        cache_attr = TransportUserMessage._CACHE_ATTRIBUTE
        msg = msg_helper.make_inbound("foo", extra="bar")
        msg.cache["cache"] = "me"
        self.assertEqual(msg[cache_attr], {"cache": "me"})

        m1 = msg_model("foo", msg=msg)
        self.assertTrue(cache_attr not in m1.msg)
        yield m1.save()

        m2 = yield msg_model.load("foo")
        self.assertTrue(cache_attr not in m2.msg)
        self.assertEqual(m2.msg, m1.msg)


class ReferencedModel(Model):
    """
    Toy model for testing fields that reference other models.
    """
    a = Integer()
    b = Unicode()


@model_field_tests
class TestForeignKey(VumiTestCase):

    class ForeignKeyModel(Model):
        """
        Toy model for ForeignKey tests.
        """
        referenced = ForeignKey(ReferencedModel, null=True)

    @needs_riak
    def test_get_data_with_foreign_key_proxy(self):
        """
        A ForeignKey field stores the referenced key in the Riak object.
        """
        referenced_model = self.manager.proxy(ReferencedModel)
        s1 = referenced_model("foo", a=5, b=u'3')
        fk_model = self.manager.proxy(self.ForeignKeyModel)

        f1 = fk_model("bar1")
        f1.referenced.set(s1)

        self.assertEqual(f1.get_data(), {
            'key': 'bar1',
            '$VERSION': None,
            'referenced': 'foo'
        })

    @needs_riak
    @Manager.calls_manager
    def test_foreignkey_fields(self):
        """
        ForeignKey fields can operate on both keys and model instances.
        """
        fk_model = self.manager.proxy(self.ForeignKeyModel)
        referenced_model = self.manager.proxy(ReferencedModel)
        s1 = referenced_model("foo", a=5, b=u'3')
        f1 = fk_model("bar")
        f1.referenced.set(s1)
        yield s1.save()
        yield f1.save()
        self.assertEqual(f1._riak_object.get_data()['referenced'], s1.key)

        f2 = yield fk_model.load("bar")
        s2 = yield f2.referenced.get()

        self.assertEqual(f2.referenced.key, "foo")
        self.assertEqual(s2.a, 5)
        self.assertEqual(s2.b, u"3")

        f2.referenced.set(None)
        s3 = yield f2.referenced.get()
        self.assertEqual(s3, None)

        f2.referenced.key = "foo"
        s4 = yield f2.referenced.get()
        self.assertEqual(s4.key, "foo")

        f2.referenced.key = None
        s5 = yield f2.referenced.get()
        self.assertEqual(s5, None)

        self.assertRaises(ValidationError, f2.referenced.set, object())

    @needs_riak
    @Manager.calls_manager
    def test_old_foreignkey_fields(self):
        """
        Old versions of the ForeignKey field relied entirely on indexes and
        didn't store the referenced key in the model data.
        """
        fk_model = self.manager.proxy(self.ForeignKeyModel)
        referenced_model = self.manager.proxy(ReferencedModel)
        s1 = referenced_model("foo", a=5, b=u'3')
        f1 = fk_model("bar")
        # Create index directly and remove data field to simulate old-style
        # index-only implementation
        f1._riak_object.add_index('referenced_bin', s1.key)
        data = f1._riak_object.get_data()
        data.pop('referenced')
        f1._riak_object.set_data(data)
        yield s1.save()
        yield f1.save()

        f2 = yield fk_model.load("bar")
        s2 = yield f2.referenced.get()

        self.assertEqual(f2.referenced.key, "foo")
        self.assertEqual(s2.a, 5)
        self.assertEqual(s2.b, u"3")

        f2.referenced.set(None)
        s3 = yield f2.referenced.get()
        self.assertEqual(s3, None)

        f2.referenced.key = "foo"
        s4 = yield f2.referenced.get()
        self.assertEqual(s4.key, "foo")

        f2.referenced.key = None
        s5 = yield f2.referenced.get()
        self.assertEqual(s5, None)

        self.assertRaises(ValidationError, f2.referenced.set, object())

    @needs_riak
    @Manager.calls_manager
    def test_reverse_foreignkey_fields(self):
        """
        When we declare a ForeignKey field, we add both a paginated index
        lookup method and a legacy non-paginated index lookup method to the
        foreign model's backlinks attribute.
        """
        fk_model = self.manager.proxy(self.ForeignKeyModel)
        referenced_model = self.manager.proxy(ReferencedModel)
        s1 = referenced_model("foo", a=5, b=u'3')
        f1 = fk_model("bar1")
        f1.referenced.set(s1)
        f2 = fk_model("bar2")
        f2.referenced.set(s1)
        yield s1.save()
        yield f1.save()
        yield f2.save()

        s2 = yield referenced_model.load("foo")
        results = yield s2.backlinks.foreignkeymodels()
        self.assertEqual(sorted(results), ["bar1", "bar2"])

        results_p1 = yield s2.backlinks.foreignkeymodel_keys()
        self.assertEqual(sorted(results_p1), ["bar1", "bar2"])
        self.assertEqual(results_p1.has_next_page(), False)


@model_field_tests
class TestManyToMany(VumiTestCase):

    @Manager.calls_manager
    def load_all_bunches_flat(self, m2m_field):
        results = []
        for result_bunch in m2m_field.load_all_bunches():
            results.extend((yield result_bunch))
        returnValue(results)

    class ManyToManyModel(Model):
        references = ManyToMany(ReferencedModel)

    @needs_riak
    def test_get_data_with_many_to_many_proxy(self):
        """
        A ManyToMany field stores the referenced keys in the Riak object.
        """
        referenced_model = self.manager.proxy(ReferencedModel)
        s1 = referenced_model("foo", a=5, b=u'3')
        mm_model = self.manager.proxy(self.ManyToManyModel)

        m1 = mm_model("bar")
        m1.references.add(s1)
        m1.save()

        self.assertEqual(m1.get_data(), {
            'key': 'bar',
            '$VERSION': None,
            'references': ['foo'],
        })

    @needs_riak
    @Manager.calls_manager
    def test_manytomany_field(self):
        """
        ManyToMany fields can operate on both keys and model instances.
        """
        mm_model = self.manager.proxy(self.ManyToManyModel)
        referenced_model = self.manager.proxy(ReferencedModel)

        s1 = referenced_model("foo", a=5, b=u'3')
        m1 = mm_model("bar")
        m1.references.add(s1)
        yield s1.save()
        yield m1.save()
        self.assertEqual(m1._riak_object.get_data()['references'], [s1.key])

        m2 = yield mm_model.load("bar")
        [s2] = yield self.load_all_bunches_flat(m2.references)

        self.assertEqual(m2.references.keys(), ["foo"])
        self.assertEqual(s2.a, 5)
        self.assertEqual(s2.b, u"3")

        m2.references.remove(s2)
        references = yield self.load_all_bunches_flat(m2.references)
        self.assertEqual(references, [])

        m2.references.add_key("foo")
        [s4] = yield self.load_all_bunches_flat(m2.references)
        self.assertEqual(s4.key, "foo")

        m2.references.remove_key("foo")
        references = yield self.load_all_bunches_flat(m2.references)
        self.assertEqual(references, [])

        self.assertRaises(ValidationError, m2.references.add, object())
        self.assertRaises(ValidationError, m2.references.remove, object())

        t1 = referenced_model("bar1", a=3, b=u'4')
        t2 = referenced_model("bar2", a=4, b=u'4')
        m2.references.add(t1)
        m2.references.add(t2)
        yield t1.save()
        yield t2.save()
        references = yield self.load_all_bunches_flat(m2.references)
        references.sort(key=lambda s: s.key)
        self.assertEqual([s.key for s in references], ["bar1", "bar2"])
        self.assertEqual(references[0].a, 3)
        self.assertEqual(references[1].a, 4)

        m2.references.clear()
        m2.references.add_key("unknown")
        self.assertEqual([], (yield self.load_all_bunches_flat(m2.references)))

    @needs_riak
    @Manager.calls_manager
    def test_old_manytomany_field(self):
        """
        Old versions of the ManyToMany field relied entirely on indexes and
        didn't store the referenced keys in the model data.
        """
        mm_model = self.manager.proxy(self.ManyToManyModel)
        referenced_model = self.manager.proxy(ReferencedModel)

        s1 = referenced_model("foo", a=5, b=u'3')
        m1 = mm_model("bar")
        # Create index directly to simulate old-style index-only implementation
        m1._riak_object.add_index('references_bin', s1.key)
        # Manually remove the entry from the data dict to allow it to be
        # set from the index value in descriptor.clean()
        data = m1._riak_object.get_data()
        data.pop('references')
        m1._riak_object.set_data(data)

        yield s1.save()
        yield m1.save()

        m2 = yield mm_model.load("bar")
        [s2] = yield self.load_all_bunches_flat(m2.references)

        self.assertEqual(m2.references.keys(), ["foo"])
        self.assertEqual(s2.a, 5)
        self.assertEqual(s2.b, u"3")

        m2.references.remove(s2)
        references = yield self.load_all_bunches_flat(m2.references)
        self.assertEqual(references, [])

        m2.references.add_key("foo")
        [s4] = yield self.load_all_bunches_flat(m2.references)
        self.assertEqual(s4.key, "foo")

        m2.references.remove_key("foo")
        references = yield self.load_all_bunches_flat(m2.references)
        self.assertEqual(references, [])

        self.assertRaises(ValidationError, m2.references.add, object())
        self.assertRaises(ValidationError, m2.references.remove, object())

        t1 = referenced_model("bar1", a=3, b=u'4')
        t2 = referenced_model("bar2", a=4, b=u'4')
        m2.references.add(t1)
        m2.references.add(t2)
        yield t1.save()
        yield t2.save()
        references = yield self.load_all_bunches_flat(m2.references)
        references.sort(key=lambda s: s.key)
        self.assertEqual([s.key for s in references], ["bar1", "bar2"])
        self.assertEqual(references[0].a, 3)
        self.assertEqual(references[1].a, 4)

        m2.references.clear()
        m2.references.add_key("unknown")
        self.assertEqual([], (yield self.load_all_bunches_flat(m2.references)))

    @needs_riak
    @Manager.calls_manager
    def test_reverse_manytomany_fields(self):
        """
        When we declare a ManyToMany field, we add both a paginated index
        lookup method and a legacy non-paginated index lookup method to the
        foreign model's backlinks attribute.
        """
        mm_model = self.manager.proxy(self.ManyToManyModel)
        referenced_model = self.manager.proxy(ReferencedModel)
        s1 = referenced_model("foo1", a=5, b=u'3')
        s2 = referenced_model("foo2", a=4, b=u'4')
        m1 = mm_model("bar1")
        m1.references.add(s1)
        m1.references.add(s2)
        m2 = mm_model("bar2")
        m2.references.add(s1)
        yield s1.save()
        yield s2.save()
        yield m1.save()
        yield m2.save()

        s1 = yield referenced_model.load("foo1")
        results = yield s1.backlinks.manytomanymodels()
        self.assertEqual(sorted(results), ["bar1", "bar2"])

        results_p1 = yield s1.backlinks.manytomanymodel_keys()
        self.assertEqual(sorted(results_p1), ["bar1", "bar2"])
        self.assertEqual(results_p1.has_next_page(), False)

        s2 = yield referenced_model.load("foo2")
        results = yield s2.backlinks.manytomanymodels()
        self.assertEqual(sorted(results), ["bar1"])

        results_p1 = yield s2.backlinks.manytomanymodel_keys()
        self.assertEqual(sorted(results_p1), ["bar1"])
        self.assertEqual(results_p1.has_next_page(), False)


@model_field_tests
class TestComputedValue(VumiTestCase):

    def test_validate_default_unicode(self):
        """
        By default, a ComputedValue field has a Unicode value.
        """
        comp = ComputedValue(lambda m: NotImplemented)
        comp.validate(u"")
        comp.validate(u"a")
        comp.validate(u"æ")
        comp.validate(u"foé")
        self.assertRaises(ValidationError, comp.validate, "")
        self.assertRaises(ValidationError, comp.validate, "foo")
        self.assertRaises(ValidationError, comp.validate, 3)

    def test_validate_listof(self):
        """
        If an explicit subtype is provided, its validation is used.
        """
        comp = ComputedValue(lambda m: NotImplemented, ListOf())
        comp.validate([u'foo', u'bar'])
        self.assertRaises(ValidationError, comp.validate,
                          u'this is not a list')
        self.assertRaises(ValidationError, comp.validate, ['a', 2])
        self.assertRaises(ValidationError, comp.validate, [1, 2])

    class ComputedValueModel(Model):
        """
        Toy model for ComputedValue tests.
        """
        a = Integer()
        b = Unicode()
        c = ListOf(Unicode())
        a_with_b = ComputedValue(
            lambda m: u"%s::%s" % (m.a, m.b), Unicode(index=True))
        b_with_a = ComputedValue(lambda m: u"%s::%s" % (m.b, m.a), Unicode())
        a_with_c = ComputedValue(
            lambda m: [u"%s::%s" % (m.a, c) for c in m.c],
            ListOf(Unicode(), index=True))

    def assert_indexes(self, mdl, indexes):
        self.assertEqual(
            mdl._riak_object.get_indexes(),
            set((k, v) for k, vs in indexes.items() for v in vs))

    @needs_riak
    @Manager.calls_manager
    def test_computedvalue_field(self):
        """
        A `ComputedValue` field gets its value from the function it's given.
        """
        ci_model = self.manager.proxy(self.ComputedValueModel)
        m1 = ci_model("foo", a=7, b=u"bar", c=[u"thing1", u"thing2"])

        # Value and index are computed at creation time.
        self.assertEqual(m1.a_with_b, u"7::bar")
        self.assertEqual(m1.b_with_a, u"bar::7")
        self.assertEqual(list(m1.a_with_c), [u"7::thing1", u"7::thing2"])
        self.assert_indexes(m1, {
            "a_with_b_bin": ["7::bar"],
            "a_with_c_bin": ["7::thing1", "7::thing2"],
        })

        # Value and index are correct after save.
        yield m1.save()
        self.assertEqual(m1.a_with_b, u"7::bar")
        self.assertEqual(m1.b_with_a, u"bar::7")
        self.assertEqual(list(m1.a_with_c), [u"7::thing1", u"7::thing2"])
        self.assert_indexes(m1, {
            "a_with_b_bin": ["7::bar"],
            "a_with_c_bin": ["7::thing1", "7::thing2"],
        })

        # Value and index are correct after load.
        m2 = yield ci_model.load("foo")
        self.assertEqual(m1.a, m2.a)
        self.assertEqual(m1.b, m2.b)
        self.assertEqual(m1.a_with_b, "7::bar")
        self.assertEqual(list(m1.a_with_c), ["7::thing1", "7::thing2"])
        self.assert_indexes(m1, {
            "a_with_b_bin": ["7::bar"],
            "a_with_c_bin": ["7::thing1", "7::thing2"],
        })

    @needs_riak
    def test_computedvalue_field_update(self):
        """
        A `ComputedValue` field's value (and index) are updated when a field
        value changes.

        This test needs a Riak manager, but never actually calls Riak.
        """
        ci_model = self.manager.proxy(self.ComputedValueModel)
        m1 = ci_model("foo", a=7, b=u"bar", c=[u"thing1", u"thing2"])

        # Value and index are computed at creation time.
        self.assertEqual(m1.a_with_b, u"7::bar")
        self.assertEqual(m1.b_with_a, u"bar::7")
        self.assertEqual(list(m1.a_with_c), [u"7::thing1", u"7::thing2"])
        self.assert_indexes(m1, {
            "a_with_b_bin": ["7::bar"],
            "a_with_c_bin": ["7::thing1", "7::thing2"],
        })

        # Value and index are recomputed after a field changes.
        m1.a = 8
        self.assertEqual(m1.a_with_b, u"8::bar")
        self.assertEqual(m1.b_with_a, u"bar::8")
        self.assertEqual(list(m1.a_with_c), [u"8::thing1", u"8::thing2"])
        self.assert_indexes(m1, {
            "a_with_b_bin": ["8::bar"],
            "a_with_c_bin": ["8::thing1", "8::thing2"],
        })

    @needs_riak
    def test_computedindex_field_cannot_set_value(self):
        """
        A `ComputedValue` field cannot have its value set manually.

        This test needs a Riak manager, but never actually calls Riak.
        """
        ci_model = self.manager.proxy(self.ComputedValueModel)
        m1 = ci_model("foo", a=7, b=u"bar")

        # Manually setting the value is impossible.
        def set_value():
            m1.a_with_b = u"a thing"
        self.assertRaises(RuntimeError, set_value)



# -*- test-case-name: vumi.transports.wechat.tests.test_message_types -*-

import re
import json
from xml.etree.ElementTree import Element, SubElement, tostring, fromstring

from vumi.transports.wechat.errors import (
    WeChatParserException, WeChatException)


def get_child_value(node, name):
    [child] = node.findall(name)
    return (child.text.strip() if child.text is not None else '')


def append(node, tag, value):
    el = SubElement(node, tag)
    el.text = value


class WeChatMessage(object):

    mandatory_fields = ()
    optional_fields = ()

    @classmethod
    def from_xml(cls, doc):
        params = [get_child_value(doc, name)
                  for name in cls.mandatory_fields]

        for field in cls.optional_fields:
            try:
                params.append(get_child_value(doc, field))
            except ValueError:
                # element not present
                continue
        return cls(*params)


class TextMessage(WeChatMessage):

    mandatory_fields = (
        'ToUserName',
        'FromUserName',
        'CreateTime',
        'Content',
    )

    optional_fields = (
        'MsgId',
    )

    def __init__(self, to_user_name, from_user_name, create_time, content,
                 msg_id=None):
        self.to_user_name = to_user_name
        self.from_user_name = from_user_name
        self.create_time = create_time
        self.content = content
        self.msg_id = msg_id

    @classmethod
    def from_vumi_message(cls, message):
        md = message['transport_metadata'].get('wechat', {})
        from_addr = md.get('ToUserName', message['from_addr'])
        return cls(message['to_addr'], from_addr,
                   message['timestamp'].strftime('%s'),
                   message['content'])

    def to_xml(self):
        xml = Element('xml')
        append(xml, 'ToUserName', self.to_user_name)
        append(xml, 'FromUserName', self.from_user_name)
        append(xml, 'CreateTime', self.create_time)
        append(xml, 'MsgType', 'text')
        append(xml, 'Content', self.content)
        return tostring(xml)

    def to_json(self):
        return json.dumps({
            'touser': self.to_user_name,
            'msgtype': 'text',
            'text': {
                'content': self.content,
            }
        })


class NewsMessage(WeChatMessage):

    # Has something URL-ish in it
    URLISH = re.compile(
        r'(?P<before>.*)'
        r'(?P<url>http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+)'
        r'(?P<after>.*?)')

    def __init__(self, to_user_name, from_user_name, create_time,
                 items=None):
        self.to_user_name = to_user_name
        self.from_user_name = from_user_name
        self.create_time = create_time
        self.items = ([] if items is None else items)

    @classmethod
    def accepts(cls, vumi_message):
        return cls.URLISH.match(vumi_message['content'])

    @classmethod
    def from_vumi_message(cls, match, vumi_message):
        md = vumi_message['transport_metadata'].get('wechat', {})
        from_addr = md.get('ToUserName', vumi_message['from_addr'])
        url_data = match.groupdict()
        return cls(
            vumi_message['to_addr'],
            from_addr,
            vumi_message['timestamp'].strftime('%s'),
            [{
                'title': '%(before)s' % url_data,
                'url': '%(url)s' % url_data,
                'description': vumi_message['content']
            }])

    def to_xml(self):
        xml = Element('xml')
        append(xml, 'ToUserName', self.to_user_name)
        append(xml, 'FromUserName', self.from_user_name)
        append(xml, 'CreateTime', self.create_time)
        append(xml, 'MsgType', 'news')
        append(xml, 'ArticleCount', str(len(self.items)))
        articles = SubElement(xml, 'Articles')
        for item in self.items:
            if not any(item.values()):
                raise WeChatException(
                    'News items must have some values.')

            item_element = SubElement(articles, 'item')
            if 'title' in item:
                append(item_element, 'Title', item['title'])
            if 'description' in item:
                append(item_element, 'Description', item['description'])
            if 'picurl' in item:
                append(item_element, 'PicUrl', item['picurl'])
            if 'url' in item:
                append(item_element, 'Url', item['url'])
        return tostring(xml)

    def to_json(self):
        return json.dumps({
            'touser': self.to_user_name,
            'msgtype': 'news',
            'news': {
                'articles': self.items
            }
        })


class EventMessage(WeChatMessage):

    mandatory_fields = (
        'ToUserName',
        'FromUserName',
        'CreateTime',
        'Event',
    )

    optional_fields = (
        'MsgId',
        'EventKey',
    )

    def __init__(self, to_user_name, from_user_name, create_time, event,
                 event_key=None):
        self.to_user_name = to_user_name
        self.from_user_name = from_user_name
        self.create_time = create_time
        self.event = event
        self.event_key = event_key


class WeChatXMLParser(object):

    ENCODING = 'utf-8'
    CLASS_MAP = {
        'text': TextMessage,
        'news': NewsMessage,
        'event': EventMessage,
    }

    @classmethod
    def parse(cls, string):
        doc = fromstring(string.decode(cls.ENCODING))
        klass = cls.get_class(doc)
        return klass.from_xml(doc)

    @classmethod
    def get_class(cls, doc):
        msg_types = doc.findall('MsgType')
        if not msg_types:
            raise WeChatParserException('No MsgType found.')

        if len(msg_types) > 1:
            raise WeChatParserException('More than 1 MsgType found.')

        [msg_type_element] = msg_types
        msg_type = msg_type_element.text.strip()
        if msg_type not in cls.CLASS_MAP:
            raise WeChatParserException(
                'Unsupported MsgType: %s' % (msg_type,))

        return cls.CLASS_MAP[msg_type]



from base64 import b64decode

from twisted.internet.defer import inlineCallbacks
from twisted.web import http

from vumi.application.http_relay import HTTPRelayApplication
from vumi.message import TransportEvent

from vumi.application.tests.helpers import ApplicationHelper
from vumi.tests.fake_connection import FakeHttpServer
from vumi.tests.helpers import VumiTestCase


class TestHTTPRelay(VumiTestCase):

    @inlineCallbacks
    def setUp(self):
        yield super(TestHTTPRelay, self).setUp()
        self.app_helper = self.add_helper(
            ApplicationHelper(HTTPRelayApplication))

    @inlineCallbacks
    def setup_resource_with_callback(self, callback):
        self.fake_http = FakeHttpServer(callback)
        self.app = yield self.app_helper.get_application({
            'url': 'http://relay.example.com/path',
            'username': 'username',
            'password': 'password',
        })
        self.app.agent_factory = self.fake_http.get_agent

    def setup_resource(self, code, content, headers):
        def handler(request):
            request.setResponseCode(code)
            for key, value in headers.items():
                request.setHeader(key, value)
            return content

        return self.setup_resource_with_callback(handler)

    @inlineCallbacks
    def test_http_relay_success_with_no_reply(self):
        yield self.setup_resource(http.OK, '', {})
        yield self.app_helper.make_dispatch_inbound("hi")
        self.assertEqual([], self.app_helper.get_dispatched_outbound())

    @inlineCallbacks
    def test_http_relay_success_with_reply_header_true(self):
        yield self.setup_resource(http.OK, 'thanks!', {
            HTTPRelayApplication.reply_header: 'true',
        })
        msg = yield self.app_helper.make_dispatch_inbound("hi")
        [response] = self.app_helper.get_dispatched_outbound()
        self.assertEqual(response['content'], 'thanks!')
        self.assertEqual(response['to_addr'], msg['from_addr'])

    @inlineCallbacks
    def test_http_relay_success_with_reply_header_false(self):
        yield self.setup_resource(http.OK, 'thanks!', {
            HTTPRelayApplication.reply_header: 'untrue!',
        })
        yield self.app_helper.make_dispatch_inbound("hi")
        self.assertEqual([], self.app_helper.get_dispatched_outbound())

    @inlineCallbacks
    def test_http_relay_success_with_bad_reply(self):
        yield self.setup_resource(http.NOT_FOUND, '', {})
        yield self.app_helper.make_dispatch_inbound("hi")
        self.assertEqual([], self.app_helper.get_dispatched_outbound())

    @inlineCallbacks
    def test_http_relay_success_with_bad_header(self):
        yield self.setup_resource(http.OK, 'thanks!', {
            'X-Other-Bad-Header': 'true',
        })
        self.assertEqual([], self.app_helper.get_dispatched_outbound())

    @inlineCallbacks
    def test_http_relay_with_basic_auth(self):
        def cb(request):
            headers = request.requestHeaders
            auth = headers.getRawHeaders('Authorization')[0]
            creds = auth.split(' ')[-1]
            username, password = b64decode(creds).split(':')
            self.assertEqual(username, 'username')
            self.assertEqual(password, 'password')
            request.setHeader(HTTPRelayApplication.reply_header, 'true')
            return 'thanks!'

        yield self.setup_resource_with_callback(cb)
        yield self.app_helper.make_dispatch_inbound("hi")
        [msg] = self.app_helper.get_dispatched_outbound()
        self.assertEqual(msg['content'], 'thanks!')

    @inlineCallbacks
    def test_http_relay_with_bad_basic_auth(self):
        def cb(request):
            request.setResponseCode(http.UNAUTHORIZED)
            return 'Not Authorized'

        yield self.setup_resource_with_callback(cb)
        yield self.app_helper.make_dispatch_inbound("hi")
        self.assertEqual([], self.app_helper.get_dispatched_outbound())

    @inlineCallbacks
    def test_http_relay_of_events(self):
        events = []

        def cb(request):
            events.append(TransportEvent.from_json(request.content.getvalue()))
            return ''

        yield self.setup_resource_with_callback(cb)
        dr = yield self.app_helper.make_dispatch_delivery_report()
        self.assertEqual([], self.app_helper.get_dispatched_outbound())
        self.assertEqual([dr], events)



import binascii

from vumi.transports.smpp.smpp_utils import unpacked_pdu_opts


def pdu_ok(pdu):
    return command_status(pdu) == 'ESME_ROK'


def seq_no(pdu):
    return pdu['header']['sequence_number']


def command_status(pdu):
    return pdu['header']['command_status']


def command_id(pdu):
    return pdu['header']['command_id']


def message_id(pdu):
    # If we have an unsuccessful response, we may not get a message_id.
    if 'body' not in pdu and not pdu_ok(pdu):
        return None
    return pdu['body']['mandatory_parameters']['message_id']


def short_message(pdu):
    return pdu['body']['mandatory_parameters']['short_message']


def pdu_tlv(pdu, tag):
    return unpacked_pdu_opts(pdu)[tag]


def chop_pdu_stream(data):
    if len(data) < 16:
        return

    bytes = binascii.b2a_hex(data[0:4])
    cmd_length = int(bytes, 16)
    if len(data) < cmd_length:
        return

    pdu, data = (data[0:cmd_length],
                 data[cmd_length:])
    return pdu, data



"""Synchronous HTTP RPC-based message transports."""

from vumi.transports.vodacom_messaging.vodacom_messaging import (
    VodacomMessagingTransport, VodacomMessagingResponse)

__all__ = ['VodacomMessagingTransport', 'VodacomMessagingResponse']



# -*- test-case-name: vumi.tests.test_servicemaker -*-
import os
import sys
import warnings

import yaml
from zope.interface import implements
from twisted.python import usage
from twisted.application.service import IServiceMaker
from twisted.plugin import IPlugin

from vumi.service import WorkerCreator
from vumi.utils import (load_class_by_string,
                        generate_worker_id)
from vumi.errors import VumiError
from vumi.sentry import SentryLoggerService


class SafeLoaderWithInclude(yaml.SafeLoader):
    def __init__(self, *args, **kwargs):
        super(SafeLoaderWithInclude, self).__init__(*args, **kwargs)
        self.add_constructor('!include', self._include)
        if isinstance(self.stream, file):
            self._root = os.path.dirname(self.stream.name)
        else:
            self._root = os.path.curdir

    def _include(self, loader, node):
        filename = os.path.join(self._root, self.construct_scalar(node))
        filename = os.path.normpath(filename)
        with open(filename) as f:
            return yaml.load(f, Loader=SafeLoaderWithInclude)


def overlay_configs(*configs):
    """Non-recursively overlay a set of configuration dictionaries"""
    config = {}

    for overlay in configs:
        config.update(overlay)

    return config


def filter_null_values(config):
    """Remove keys with None values from a dictionary."""
    return dict(item for item in config.iteritems() if item[1] is not None)


def read_yaml_config(config_file, optional=True):
    """Parse an (usually) optional YAML config file."""
    if optional and config_file is None:
        return {}
    with file(config_file, 'r') as stream:
        # Assume we get a dict out of this.
        return yaml.load(stream, Loader=SafeLoaderWithInclude)


class VumiOptions(usage.Options):
    """
    Options global to everything vumi.
    """
    optParameters = [
        ["hostname", None, None, "AMQP broker (*)"],
        ["port", None, None, "AMQP port (*)", int],
        ["username", None, None, "AMQP username (*)"],
        ["password", None, None, "AMQP password (*)"],
        ["vhost", None, None, "AMQP virtual host (*)"],
        ["specfile", None, None, "AMQP spec file (*)"],
        ["sentry", None, None, "Sentry DSN (*)"],
        ["vumi-config", None, None,
         "YAML config file for setting core vumi options (any command-line"
         " parameter marked with an asterisk)"],
        ["system-id", None, None,
         "An identifier for a collection of Vumi workers"],
    ]

    default_vumi_options = {
        "hostname": "127.0.0.1",
        "port": 5672,
        "username": "vumi",
        "password": "vumi",
        "vhost": "/develop",
        "specfile": "amqp-spec-0-8.xml",
        "sentry": None,
        }

    def get_vumi_options(self):
        # We don't want these to get in the way later.
        vumi_option_params = {}
        for opt in (i[0] for i in VumiOptions.optParameters):
            vumi_option_params[opt] = self.pop(opt)

        config_file = vumi_option_params.pop('vumi-config')

        # non-recursive overlay is safe because vumi options are
        # all simple key-value pairs
        return overlay_configs(
            self.default_vumi_options,
            read_yaml_config(config_file),
            filter_null_values(vumi_option_params))

    def postOptions(self):
        self.vumi_options = self.get_vumi_options()


class StartWorkerOptions(VumiOptions):
    """
    Options to the vumi_worker twistd plugin.
    """

    optFlags = [
        ["worker-help", None,
         "Print out a usage message for the worker-class and exit"],
        ]

    optParameters = [
        ["worker-class", None, None, "Class of a worker to start"],
        ["worker_class", None, None, "Deprecated. See --worker-class instead"],
        ["config", None, None, "YAML config file for worker configuration"
         " options"],
        ["maxthreads", None, None, "Maximum size of reactor thread pool", int],
    ]

    longdesc = """Launch an instance of a vumi worker process."""

    def __init__(self):
        VumiOptions.__init__(self)
        self.set_options = {}

    def opt_set_option(self, keyvalue):
        """Set a worker configuration option (overrides values
        specified in the file passed to --config)."""
        key, _sep, value = keyvalue.partition(':')
        self.set_options[key] = value

    def exit(self):
        # So we can stub it out in tests.
        sys.exit(0)

    def emit(self, text):
        # So we can stub it out in tests.
        print text

    def do_worker_help(self):
        """Print out a usage message for the worker-class and exit"""
        worker_class = load_class_by_string(self.worker_class)
        self.emit(worker_class.__doc__)
        config_class = getattr(worker_class, 'CONFIG_CLASS', None)
        if config_class is not None:
            self.emit(config_class.__doc__)
        self.emit("")
        self.exit()

    def get_worker_class(self):
        worker_class = self.opts.pop('worker-class')
        depr_worker_class = self.opts.pop('worker_class')

        if depr_worker_class is not None:
            warnings.warn("The --worker_class option is deprecated since"
                          " Vumi 0.3. Please use --worker-class instead.",
                          category=DeprecationWarning)
            if worker_class is None:
                worker_class = depr_worker_class

        if worker_class is None:
            raise VumiError("please specify --worker-class")

        return worker_class

    def get_worker_config(self):
        config_file = self.opts.pop('config')

        # non-recursive overlay is safe because set_options
        # can only contain simple key-value pairs.
        return overlay_configs(
            read_yaml_config(config_file),
            self.set_options)

    def get_maxthreads(self):
        return self.opts.pop("maxthreads")

    def postOptions(self):
        VumiOptions.postOptions(self)

        self.worker_class = self.get_worker_class()

        if self.opts.pop('worker-help'):
            self.do_worker_help()

        self.worker_config = self.get_worker_config()

        self.maxthreads = self.get_maxthreads()


class VumiWorkerServiceMaker(object):
    implements(IServiceMaker, IPlugin)
    # the name of our plugin, this will be the subcommand for twistd
    # e.g. $ twistd -n vumi_worker --option1= ...
    tapname = "vumi_worker"
    # description, also for twistd
    description = "Start a Vumi worker"
    # what command line options does this service expose
    options = StartWorkerOptions

    def set_maxthreads(self, maxthreads):
        from twisted.internet import reactor

        if maxthreads is not None:
            reactor.suggestThreadPoolSize(maxthreads)

    def makeService(self, options):
        sentry_dsn = options.vumi_options.pop('sentry', None)
        class_name = options.worker_class.rpartition('.')[2].lower()
        logger_name = options.worker_config.get('worker_name', class_name)
        system_id = options.vumi_options.get('system-id', 'global')
        worker_id = generate_worker_id(system_id, logger_name)

        self.set_maxthreads(options.maxthreads)

        worker_creator = WorkerCreator(options.vumi_options)
        worker = worker_creator.create_worker(options.worker_class,
                                              options.worker_config)

        if sentry_dsn is not None:
            sentry_service = SentryLoggerService(sentry_dsn,
                                                 logger_name,
                                                 worker_id)
            worker.addService(sentry_service)

        return worker


class DeprecatedStartWorkerServiceMaker(VumiWorkerServiceMaker):
    tapname = "start_worker"
    description = "Deprecated copy of vumi_worker. Use vumi_worker instead."



# -*- test-case-name: vumi.scripts.tests.test_benchmark_persist -*-
import sys
import time
from twisted.python import usage
from twisted.internet import reactor
from twisted.internet.defer import maybeDeferred, inlineCallbacks, DeferredList

from vumi.message import TransportUserMessage
from vumi.persist.model import Model
from vumi.persist.txriak_manager import TxRiakManager
from vumi.persist.fields import VumiMessage


class Options(usage.Options):
    optParameters = [
        ["messages", "m", "1000",
         "Total number of messages to write and read back."],
        ["concurrent-messages", "c", "100",
         "Number of messages to read and write concurrently"],
    ]

    longdesc = """Benchmarks vumi.persist.model.Model"""


class MessageModel(Model):
    msg = VumiMessage(TransportUserMessage)


class WriteReadBenchmark(object):
    """
    Writes messages to Riak and then reads them back.
    """

    def __init__(self, options):
        self.messages = int(options['messages'])
        self.concurrent = int(options['concurrent-messages'])

    def make_batches(self):
        num_batches, rem = divmod(self.messages, self.concurrent)
        batches = [self.make_batch(i, self.concurrent)
                   for i in range(num_batches)]
        if rem:
            batches.append(self.make_batch(num_batches, rem))
        return batches

    def make_batch(self, batch_no, num_msgs):
        return [TransportUserMessage(to_addr="1234", from_addr="5678",
                    transport_name="bench", transport_type="sms",
                    content="Batch: %d. Msg: %d" % (batch_no, i))
                for i in range(num_msgs)]

    def write_batch(self, model, msgs):
        print "  Writing %d messages." % len(msgs)
        deferreds = []
        for msg in msgs:
            msg_obj = model(key=msg['message_id'], msg=msg)
            deferreds.append(msg_obj.save())
        return DeferredList(deferreds)

    def read_batch(self, model, msgs):
        print "  Reading %d messages." % len(msgs)
        deferreds = []
        for msg in msgs:
            deferreds.append(model.load(msg['message_id']))
        return DeferredList(deferreds)

    @inlineCallbacks
    def run(self):
        manager = TxRiakManager.from_config({'bucket_prefix': 'test.bench.'})
        model = manager.proxy(MessageModel)
        yield manager.purge_all()

        msg_batches = self.make_batches()

        start = time.time()

        for batch in msg_batches:
            yield self.write_batch(model, batch)

        write_done = time.time()
        write_time = write_done - start
        print "Write took %.2f seconds (%.2f msgs/s)" % (
                write_time, self.messages / write_time)

        result_batches = []
        for batch in msg_batches:
            r = yield self.read_batch(model, batch)
            result_batches.append(r)

        read_done = time.time()
        read_time = read_done - write_done
        print "Read took %.2f seconds (%.2f msgs/s)" % (
                read_time, self.messages / read_time)

        for batch, result_batch in zip(msg_batches, result_batches):
            for msg, (good, stored_msg) in zip(batch, result_batch):
                if not good or stored_msg is None:
                    raise RuntimeError("Failed to retrieve message (%s)"
                                       % msg['content'])
                if not(msg == stored_msg.msg):  # TODO: fix message !=
                    raise RuntimeError("Message %r does not equal stored"
                                       " message %r" % (msg, stored_msg.msg))

        print "Messages retrieved successfully."

        yield manager.purge_all()
        print "Messages purged."

if __name__ == '__main__':
    try:
        options = Options()
        options.parseOptions()
    except usage.UsageError, errortext:
        print '%s: %s' % (sys.argv[0], errortext)
        print '%s: Try --help for usage details.' % (sys.argv[0])
        sys.exit(1)

    bench = WriteReadBenchmark(options)

    def _eb(f):
        f.printTraceback()

    def _main():
        d = maybeDeferred(bench.run)
        d.addErrback(_eb)
        d.addBoth(lambda _: reactor.stop())

    reactor.callLater(0, _main)
    reactor.run()



# -*- test-case-name: vumi.application.tests.test_http_relay -*-

from base64 import b64encode

from twisted.python import log
from twisted.web import http
from twisted.internet.defer import inlineCallbacks

from vumi.application.base import ApplicationWorker
from vumi.utils import http_request_full
from vumi.errors import VumiError
from vumi.config import ConfigText, ConfigUrl


class HTTPRelayError(VumiError):
    pass


class HTTPRelayConfig(ApplicationWorker.CONFIG_CLASS):

    # TODO: Make these less static?
    url = ConfigUrl(
        "URL to submit incoming message to.", required=True, static=True)
    event_url = ConfigUrl(
        "URL to submit incoming events to. (Defaults to the same as 'url').",
        static=True)
    http_method = ConfigText(
        "HTTP method for submitting messages.", default='POST', static=True)
    auth_method = ConfigText(
        "HTTP authentication method.", default='basic', static=True)

    username = ConfigText("Username for HTTP authentication.", default='')
    password = ConfigText("Password for HTTP authentication.", default='')


class HTTPRelayApplication(ApplicationWorker):
    CONFIG_CLASS = HTTPRelayConfig

    reply_header = 'X-Vumi-HTTPRelay-Reply'
    agent_factory = None  # For swapping out the Agent we use in tests.

    def validate_config(self):
        self.supported_auth_methods = {
            'basic': self.generate_basic_auth_headers,
        }
        # XXX: Is this the best way to do this?
        if 'event_url' not in self.config:
            self.config['event_url'] = self.config['url']
        config = self.get_static_config()
        if config.auth_method not in self.supported_auth_methods:
            raise HTTPRelayError(
                'HTTP Authentication method %s not supported' % (
                    repr(config.auth_method,)))

    def generate_basic_auth_headers(self, username, password):
        credentials = ':'.join([username, password])
        auth_string = b64encode(credentials.encode('utf-8'))
        return {
            'Authorization': ['Basic %s' % (auth_string,)]
        }

    def get_auth_headers(self, config):
        if config.username:
            handler = self.supported_auth_methods.get(config.auth_method)
            return handler(config.username, config.password)
        return {}

    @inlineCallbacks
    def consume_user_message(self, message):
        config = yield self.get_config(message)
        headers = self.get_auth_headers(config)
        response = yield http_request_full(
            config.url.geturl(), message.to_json(), headers,
            config.http_method, agent_class=self.agent_factory)
        headers = response.headers
        if response.code == http.OK:
            if headers.hasHeader(self.reply_header):
                raw_headers = headers.getRawHeaders(self.reply_header)
                content = response.delivered_body.strip()
                if (raw_headers[0].lower() == 'true') and content:
                    self.reply_to(message, content)
        else:
            log.err('%s responded with %s' % (
                config.url.geturl(), response.code))

    @inlineCallbacks
    def relay_event(self, event):
        config = yield self.get_config(event)
        headers = self.get_auth_headers(config)
        yield http_request_full(
            config.event_url.geturl(), event.to_json(), headers,
            config.http_method, agent_class=self.agent_factory)

    @inlineCallbacks
    def consume_ack(self, event):
        yield self.relay_event(event)

    @inlineCallbacks
    def consume_delivery_report(self, event):
        yield self.relay_event(event)



# -*- test-case-name: vumi.tests.test_testutils -*-

import re
from datetime import datetime, timedelta
import warnings
from functools import wraps

import pytz
from twisted.internet import reactor
from twisted.web.resource import Resource
from twisted.internet.defer import DeferredQueue, inlineCallbacks
from twisted.python import log

from vumi.service import get_spec, WorkerCreator
from vumi.tests.fake_amqp import FakeAMQClient
from vumi.utils import vumi_resource_path, flatten_generator, LogFilterSite

# For backcompat:
from vumi.tests.helpers import import_filter, import_skip
import_filter, import_skip  # To keep pyflakes happy.


class UTCNearNow(object):
    def __init__(self, offset=10):
        self.now = datetime.utcnow()
        self.utcnow = self.now.replace(tzinfo=pytz.UTC)
        self.offset = timedelta(offset)

    def __eq__(self, other):
        now = self.now
        if other.tzinfo:
            now = self.utcnow
        return (now - self.offset) < other < (now + self.offset)


class RegexMatcher(object):
    def __init__(self, regex):
        self.regex = re.compile(regex)

    def __eq__(self, other):
        return self.regex.match(other)


def get_fake_amq_client(broker=None):
    spec = get_spec(vumi_resource_path("amqp-spec-0-8.xml"))
    return FakeAMQClient(spec, {}, broker)


def get_stubbed_worker(worker_class, config=None, broker=None):
    worker = worker_class({}, config)
    worker._amqp_client = get_fake_amq_client(broker)
    return worker


class StubbedWorkerCreator(WorkerCreator):
    broker = None

    def _connect(self, worker, timeout, bindAddress):
        amq_client = get_fake_amq_client(self.broker)
        self.broker = amq_client.broker  # So we use the same broker for all.
        reactor.callLater(0, worker._amqp_connected, amq_client)


def FakeRedis():
    warnings.warn("Use of FakeRedis is deprecated. "
                  "Use persist.tests.fake_redis instead.",
                  category=DeprecationWarning)
    from vumi.persist import fake_redis
    return fake_redis.FakeRedis()


class LogCatcher(object):
    """Context manager for gathering logs in tests.

    :param str system:
        Only log events whose 'system' value contains the given
        regular expression pattern will be gathered. Default: None
        (i.e. keep all log events).

    :param str message:
        Only log events whose message contains the given regular
        expression pattern will be gathered. The message is
        constructed by joining the elements in the 'message' value
        with a space (the same way Twisted does). Default: None
        (i.e. keep all log events).

    :param int log_level:
        Only log events whose logLevel is equal to the given level
        will be gathered. Default: None (i.e. keep all log events).
    """

    def __init__(self, system=None, message=None, log_level=None):
        self.logs = []
        self.system = re.compile(system) if system is not None else None
        self.message = re.compile(message) if message is not None else None
        self.log_level = log_level

    @property
    def errors(self):
        return [ev for ev in self.logs if ev["isError"]]

    def messages(self):
        return [" ".join(msg['message']) for msg in self.logs
                if not msg["isError"]]

    def _keep_log(self, event_dict):
        if self.system is not None:
            if not self.system.search(event_dict.get('system', '-')):
                return False
        if self.message is not None:
            log_message = " ".join(event_dict.get('message', []))
            if not self.message.search(log_message):
                return False
        if self.log_level is not None:
            if event_dict.get('logLevel', None) != self.log_level:
                return False
        return True

    def _gather_logs(self, event_dict):
        if self._keep_log(event_dict):
            self.logs.append(event_dict)

    def __enter__(self):
        log.theLogPublisher.addObserver(self._gather_logs)
        return self

    def __exit__(self, *exc_info):
        log.theLogPublisher.removeObserver(self._gather_logs)


class MockResource(Resource):
    isLeaf = True

    def __init__(self, handler):
        Resource.__init__(self)
        self.handler = handler

    def render_GET(self, request):
        return self.handler(request)

    def render_POST(self, request):
        return self.handler(request)

    def render_PUT(self, request):
        return self.handler(request)


class MockHttpServer(object):
    """
    NOTE: This is deprecated.
          Please use :class:`vumi.tests.http_helpers.MockHttpHelper` instead.
    """

    def __init__(self, handler=None):
        self.queue = DeferredQueue()
        self._handler = handler or self.handle_request
        self._webserver = None
        self.addr = None
        self.url = None

    def handle_request(self, request):
        self.queue.put(request)

    @inlineCallbacks
    def start(self):
        root = MockResource(self._handler)
        site_factory = LogFilterSite(root)
        self._webserver = yield reactor.listenTCP(
            0, site_factory, interface='127.0.0.1')
        self.addr = self._webserver.getHost()
        self.url = "http://%s:%s/" % (self.addr.host, self.addr.port)

    @inlineCallbacks
    def stop(self):
        yield self._webserver.stopListening()
        yield self._webserver.loseConnection()


def maybe_async(sync_attr):
    """Decorate a method that may be sync or async.

    This redecorates with the either @inlineCallbacks or @flatten_generator,
    depending on the `sync_attr`.
    """
    if callable(sync_attr):
        # If we don't get a sync attribute name, default to 'is_sync'.
        return maybe_async('is_sync')(sync_attr)

    def redecorate(func):
        @wraps(func)
        def wrapper(self, *args, **kw):
            if getattr(self, sync_attr):
                return flatten_generator(func)(self, *args, **kw)
            return inlineCallbacks(func)(self, *args, **kw)
        return wrapper

    return redecorate



from vumi.config import (
    ConfigText, ConfigInt, ConfigBool, ConfigClientEndpoint, ConfigDict,
    ConfigFloat, ConfigClassName, ClientEndpointFallback)
from vumi.transports.smpp.iprocessors import (
    IDeliveryReportProcessor, IDeliverShortMessageProcessor,
    ISubmitShortMessageProcessor)
from vumi.codecs.ivumi_codecs import IVumiCodec
from vumi.transports.base import Transport


class SmppTransportConfig(Transport.CONFIG_CLASS):

    twisted_endpoint = ConfigClientEndpoint(
        'The SMPP endpoint to connect to.',
        required=True, static=True,
        fallbacks=[ClientEndpointFallback()])
    initial_reconnect_delay = ConfigInt(
        'How long (in seconds) to wait between reconnecting attempts. '
        'Defaults to 5 seconds.', default=5, static=True)
    throttle_delay = ConfigFloat(
        "Delay (in seconds) before retrying a message after receiving "
        "`ESME_RTHROTTLED` or `ESME_RMSGQFUL`.", default=0.1, static=True)
    deliver_sm_decoding_error = ConfigText(
        'The error to respond with when we were unable to decode all parts '
        'of a PDU.', default='ESME_RDELIVERYFAILURE', static=True)
    submit_sm_expiry = ConfigInt(
        'How long (in seconds) to wait for the SMSC to return with a '
        '`submit_sm_resp`. Defaults to 24 hours.',
        default=(60 * 60 * 24), static=True)
    disable_ack = ConfigBool(
        'Disable publishing of `ack` events. In some cases this event '
        'causes more noise than signal. It can optionally be turned off. '
        'Defaults to False.',
        default=False, static=True)
    third_party_id_expiry = ConfigInt(
        'How long (in seconds) to keep 3rd party message IDs around to allow '
        'for matching submit_sm_resp and delivery report messages. Defaults '
        'to 1 week.',
        default=(60 * 60 * 24 * 7), static=True)
    final_dr_third_party_id_expiry = ConfigInt(
        'How long (in seconds) to keep 3rd party message IDs around after '
        'receiving a success or failure delivery report for the message.',
        default=(60 * 60), static=True)
    completed_multipart_info_expiry = ConfigInt(
        'How long (in seconds) to keep multipart message info for completed '
        'multipart messages around to avoid pending operations accidentally '
        'recreating them without an expiry time. Defaults to 1 hour.',
        default=(60 * 60), static=True)
    redis_manager = ConfigDict(
        'How to connect to Redis.', default={}, static=True)
    split_bind_prefix = ConfigText(
        "This is the Redis prefix to use for storing things like sequence "
        "numbers and message ids for delivery report handling. It defaults "
        "to `<system_id>@<transport_name>`. "
        "This should *ONLY* be done for TX & RX since messages sent via "
        "the TX bind are handled by the RX bind and they need to share the "
        "same prefix for the lookup for message ids in delivery reports to "
        "work.", default='', static=True)
    codec_class = ConfigClassName(
        'Which class should be used to handle character encoding/decoding. '
        'MUST implement `IVumiCodec`.',
        default='vumi.codecs.VumiCodec',
        static=True, implements=IVumiCodec)
    delivery_report_processor = ConfigClassName(
        'Which delivery report processor to use. '
        'MUST implement `IDeliveryReportProcessor`.',
        default=('vumi.transports.smpp.processors.'
                 'DeliveryReportProcessor'),
        static=True, implements=IDeliveryReportProcessor)
    delivery_report_processor_config = ConfigDict(
        'The configuration for the `delivery_report_processor`.',
        default={}, static=True)
    deliver_short_message_processor = ConfigClassName(
        'Which deliver short message processor to use. '
        'MUST implement `IDeliverShortMessageProcessor`.',
        default='vumi.transports.smpp.processors.DeliverShortMessageProcessor',
        static=True, implements=IDeliverShortMessageProcessor)
    deliver_short_message_processor_config = ConfigDict(
        'The configuration for the `deliver_short_message_processor`.',
        default={}, static=True)
    submit_short_message_processor = ConfigClassName(
        'Which submit short message processor to use. '
        'Should implements `ISubmitShortMessageProcessor`.',
        default='vumi.transports.smpp.processors.SubmitShortMessageProcessor',
        static=True, implements=ISubmitShortMessageProcessor)
    submit_short_message_processor_config = ConfigDict(
        'The configuration for the `submit_short_message_processor`.',
        default={}, static=True)
    system_id = ConfigText(
        'User id used to connect to the SMPP server.', required=True,
        static=True)
    password = ConfigText(
        'Password for the system id.', required=True, static=True)
    system_type = ConfigText(
        "Additional system metadata that is passed through to the SMPP "
        "server on connect.", default="", static=True)
    interface_version = ConfigText(
        "SMPP protocol version. Default is '34' (i.e. version 3.4).",
        default="34", static=True)
    service_type = ConfigText(
        'The SMPP service type.', default="", static=True)
    address_range = ConfigText(
        "Address range to receive. (SMSC-specific format, default empty.)",
        default="", static=True)
    dest_addr_ton = ConfigInt(
        'Destination TON (type of number).', default=0, static=True)
    dest_addr_npi = ConfigInt(
        'Destination NPI (number plan identifier). '
        'Default 1 (ISDN/E.164/E.163).', default=1, static=True)
    source_addr_ton = ConfigInt(
        'Source TON (type of number).', default=0, static=True)
    source_addr_npi = ConfigInt(
        'Source NPI (number plan identifier).', default=0, static=True)
    registered_delivery = ConfigBool(
        'Whether or not to request delivery reports. Default True.',
        default=True, static=True)
    smpp_bind_timeout = ConfigInt(
        'How long (in seconds) to wait for a succesful bind. Default 30.',
        default=30, static=True)
    smpp_enquire_link_interval = ConfigInt(
        "How long (in seconds) to delay before reconnecting to the server "
        "after being disconnected. Some WASPs, e.g. Clickatell require a 30s "
        "delay before reconnecting. In these cases a 45s "
        "`initial_reconnect_delay` is recommended. Default 55.",
        default=55, static=True)
    mt_tps = ConfigInt(
        'Mobile Terminated Transactions per Second. The Maximum Vumi '
        'messages per second to attempt to put on the wire. '
        'Defaults to 0 which means no throttling is applied. '
        '(NOTE: 1 Vumi message may result in multiple PDUs)',
        default=0, static=True, required=False)

    # TODO: Deprecate these fields when confmodel#5 is done.
    host = ConfigText(
        "*DEPRECATED* 'host' and 'port' fields may be used in place of the"
        " 'twisted_endpoint' field.", static=True)
    port = ConfigInt(
        "*DEPRECATED* 'host' and 'port' fields may be used in place of the"
        " 'twisted_endpoint' field.", static=True)



# -*- test-case-name: vumi.middleware.tests.test_logging -*-

from confmodel.fields import ConfigText

from vumi.middleware import BaseMiddleware
from vumi.middleware.base import BaseMiddlewareConfig
from vumi import log


class LoggingMiddlewareConfig(BaseMiddlewareConfig):
    """
    Configuration class for the logging middleware.
    """
    log_level = ConfigText(
        "Log level from :mod:`vumi.log` to log inbound and outbound messages "
        "and events at", default='info', static=True)
    failure_log_level = ConfigText(
        "Log level from :mod:`vumi.log` to log failure messages at",
        default='error', static=True)


class LoggingMiddleware(BaseMiddleware):
    """Middleware for logging messages published and consumed by
    transports and applications.

    Optional configuration:

    :param string log_level:
        Log level from :mod:`vumi.log` to log inbound and outbound
        messages and events at. Default is `info`.
    :param string failure_log_level:
        Log level from :mod:`vumi.log` to log failure messages at.
        Default is `error`.
    """
    CONFIG_CLASS = LoggingMiddlewareConfig

    def setup_middleware(self):
        log_level = self.config.log_level
        self.message_logger = getattr(log, log_level)
        failure_log_level = self.config.failure_log_level
        self.failure_logger = getattr(log, failure_log_level)

    def _log(self, direction, logger, msg, connector_name):
        logger("Processed %s message for %s: %s" % (
                direction, connector_name, msg.to_json()))
        return msg

    def handle_inbound(self, message, connector_name):
        return self._log(
            "inbound", self.message_logger, message, connector_name)

    def handle_outbound(self, message, connector_name):
        return self._log(
            "outbound", self.message_logger, message, connector_name)

    def handle_event(self, event, connector_name):
        return self._log("event", self.message_logger, event, connector_name)

    def handle_failure(self, failure, connector_name):
        return self._log(
            "failure", self.failure_logger, failure, connector_name)



"""Tests for vumi.application.session."""

import time

from vumi.persist.fake_redis import FakeRedis
from vumi.application import SessionManager
from vumi.tests.helpers import VumiTestCase


class TestSessionManager(VumiTestCase):
    def setUp(self):
        self.fake_redis = FakeRedis()
        self.add_cleanup(self.fake_redis.teardown)
        self.sm = SessionManager(self.fake_redis, prefix="test")
        self.add_cleanup(self.sm.stop)

    def test_active_sessions(self):
        def get_sessions():
            return sorted(self.sm.active_sessions())

        def ids():
            return [x[0] for x in get_sessions()]

        self.assertEqual(ids(), [])
        self.sm.create_session("u1")
        self.assertEqual(ids(), ["u1"])
         # 10 seconds later
        self.sm.create_session("u2", created_at=time.time() + 10)
        self.assertEqual(ids(), ["u1", "u2"])

        s1, s2 = get_sessions()
        self.assertTrue(s1[1]['created_at'] < s2[1]['created_at'])

    def test_schedule_session_expiry(self):
        self.sm.max_session_length = 60.0
        self.sm.create_session("u1")

    def test_create_and_retrieve_session(self):
        session = self.sm.create_session("u1")
        self.assertEqual(sorted(session.keys()), ['created_at'])
        self.assertTrue(time.time() - float(session['created_at']) < 10.0)
        loaded = self.sm.load_session("u1")
        self.assertEqual(loaded, session)

    def test_save_session(self):
        test_session = {"foo": 5, "bar": "baz"}
        self.sm.create_session("u1")
        self.sm.save_session("u1", test_session)
        session = self.sm.load_session("u1")
        self.assertTrue(session.pop('created_at') is not None)
        # Redis saves & returns all session values as strings
        self.assertEqual(session, dict([map(str, kvs) for kvs
                                        in test_session.items()]))

    def test_lazy_clearing(self):
        self.sm.save_session('user_id', {})
        self.assertEqual(list(self.sm.active_sessions()), [])
